import os
import re
import shutil

from ..utils.compare_templates import compare_zip_code_folders


def checkTemplateForVersionUpdate(template_path):
    if os.path.exists(template_path):
        # print("Template already exist")
        templateIsSame = compare_zip_code_folders(f"{template_path}.zip",os.path.join(template_path, "lts.zip"))

        if templateIsSame:
            os.remove(f"{template_path}.zip") #template did not change, so delete extracted feature
        else:
            next_version = get_next_version_file(template_path)

            #rename old lts version
            last_version_name = template_path + "/" + next_version #new name for last version (current lts version)
            os.rename(os.path.join(template_path, "lts.zip"),last_version_name)

            #move update template to folder
            shutil.move(f"{template_path}.zip", os.path.join(template_path, "lts.zip"))

    else:
        # print("Fresh template")
        # make template base folder
        os.makedirs(template_path, exist_ok=True)

        # move extracted to new folder and rename to lts.zip
        shutil.move(f"{template_path}.zip", os.path.join(template_path, "lts.zip"))


def get_next_version_file(folder_path):

    version_pattern = re.compile(r'^(\d+(?:\.\d+)?)\.zip$')
    versions = []

    for fname in os.listdir(folder_path):
        if fname == "lts.zip":
            continue
        match = version_pattern.match(fname)
        if match:
            ver_str = match.group(1)
            parts = ver_str.split('.')
            if len(parts) == 1:
                major = int(parts[0])
                minor = 0
            else:
                major = int(parts[0])
                minor = int(parts[1])
            versions.append((major, minor))

    if not versions:
        return "0.1.zip"

    # Find highest version
    versions.sort()
    major, minor = versions[-1]

    # Calculate next version
    if minor < 9:
        next_major = major
        next_minor = minor + 1
    else:
        next_major = major + 1
        next_minor = 0

    next_version = f"{next_major}.{next_minor}.zip" if next_minor > 0 else f"{next_major}.zip"
    return next_version


import re
import ast
import ast

from ....parsers.python.parser import PythonBlockParser
from .var_utils import get_assigned_variables, removeDecorator

pattern = r'\b(?:\w+\.)+\w+[,]?\b'


parser = PythonBlockParser()

def get_referenced_variables(block, root=False):
    """Extracts variable names referenced/used within the given block using AST."""
    referenced_vars = set()
    assigned_vars = set()  # Track assignments to exclude them from references
    
    block = removeDecorator(block)  # Assuming you have this function
    
    try:
        tree = ast.parse(block.strip())
        
        # First pass: collect all assigned variables
        for node in ast.walk(tree):
            if isinstance(node, ast.Assign):
                for target in node.targets:
                    if isinstance(target, ast.Name):
                        assigned_vars.add(target.id)
            elif isinstance(node, ast.AnnAssign):
                if isinstance(node.target, ast.Name):
                    assigned_vars.add(node.target.id)
            elif isinstance(node, ast.FunctionDef):
                assigned_vars.add(node.name)
                # Add function parameters as assigned
                for arg in node.args.args:
                    assigned_vars.add(arg.arg)
            elif isinstance(node, ast.ClassDef):
                assigned_vars.add(node.name)
        
        # Second pass: collect all referenced variables
        for node in ast.walk(tree):
            if isinstance(node, ast.Name) and isinstance(node.ctx, ast.Load):
                # Only include variables that are being loaded/referenced
                # and are not built-ins or assigned within this block
                var_name = node.id
                if (var_name not in assigned_vars and 
                    var_name not in dir(__builtins__) and
                    not var_name.startswith('__')):
                    referenced_vars.add(var_name)
            
            elif isinstance(node, ast.Attribute):
                # Handle attribute access (e.g., obj.method, module.function)
                if isinstance(node.value, ast.Name):
                    var_name = node.value.id
                    if (var_name not in assigned_vars and 
                        var_name not in dir(__builtins__) and
                        not var_name.startswith('__')):
                        referenced_vars.add(var_name)
            
            elif isinstance(node, ast.Call):
                # Handle function calls
                if isinstance(node.func, ast.Name):
                    func_name = node.func.id
                    if (func_name not in assigned_vars and 
                        func_name not in dir(__builtins__) and
                        not func_name.startswith('__')):
                        referenced_vars.add(func_name)
        
        if root and len(referenced_vars) > 0:
            return list(referenced_vars)[0]
            
    except:
        pass  # Silently handle parsing errors like the original
    
    return referenced_vars


def get_all_variables_and_parameters(block):
    """
    Comprehensive function that extracts both assigned and referenced variables,
    including function parameters and their usage.
    """
    all_vars = {
        'assigned': set(),
        'referenced': set(), 
        'parameters': set(),
        'function_calls': set(),
        'attributes': set()
    }
    
    block = removeDecorator(block)
    
    try:
        tree = ast.parse(block.strip())
        
        for node in ast.walk(tree):
            # Assignments
            if isinstance(node, ast.Assign):
                for target in node.targets:
                    if isinstance(target, ast.Name):
                        all_vars['assigned'].add(target.id)
                        
            elif isinstance(node, ast.AnnAssign):
                if isinstance(node.target, ast.Name):
                    all_vars['assigned'].add(node.target.id)
                    
            # Function definitions and parameters
            elif isinstance(node, ast.FunctionDef):
                all_vars['assigned'].add(node.name)
                for arg in node.args.args:
                    all_vars['parameters'].add(arg.arg)
                    
            elif isinstance(node, ast.ClassDef):
                all_vars['assigned'].add(node.name)
                
            # Variable references
            elif isinstance(node, ast.Name) and isinstance(node.ctx, ast.Load):
                var_name = node.id
                if (not var_name.startswith('__') and 
                    var_name not in dir(__builtins__)):
                    all_vars['referenced'].add(var_name)
                    
            # Function calls
            elif isinstance(node, ast.Call):
                if isinstance(node.func, ast.Name):
                    all_vars['function_calls'].add(node.func.id)
                elif isinstance(node.func, ast.Attribute):
                    if isinstance(node.func.value, ast.Name):
                        all_vars['referenced'].add(node.func.value.id)
                        
            # Attribute access
            elif isinstance(node, ast.Attribute):
                if isinstance(node.value, ast.Name):
                    all_vars['attributes'].add(f"{node.value.id}.{node.attr}")
                    all_vars['referenced'].add(node.value.id)
    
    except:
        pass
    
    return all_vars

def get_refrenced_variables_from_file(feature, blocks_name):
    refrences = get_referenced_variables(feature)
    ref_list = []
    for i in refrences:
        if i in blocks_name:
            ref_list.append({"packagePath":".","imports":i})
            
    return ref_list



def reformatLine(line, project_name):
    line = line.split(".")
    feature = line.pop()
    file = line.pop()

    if file == project_name:
        return ["/".join(line) + "/" + file + "/" + feature + ".py", "__all__"]
    else:
        file += ".py"

    return ["/".join(line) + "/" + file, feature]

def getFileDependencies(code,project_name):
    matches = re.findall(pattern, code)
    dependencies = []

    for i in matches:
        dependencies.append(reformatLine(i,project_name))

    return dependencies

def extract_words_from_code(code):
    if code == None:
        return set()
    # Remove strings and comments
    code = re.sub(r'(".*?"|".*?")', '', code)  # Remove strings
    code = re.sub(r'#.*', '', code)  # Remove comments
    
    # Extract words using regex (identifiers, keywords, function names, variable names)
    words = re.findall(r'[a-zA-Z_][a-zA-Z0-9_]*', code)
    
    return set(words)

def getBlockDependencies(block, all_blocks):
    importLine = []
    block_words = extract_words_from_code(block)

    # Get assigned variables in the current block
    assigned_vars = get_assigned_variables(block)

    # Remove assigned variables from the dependencies list
    filtered_words = [word for word in block_words if word not in assigned_vars]
    
    filtered_chunks = []
    other_chunks = [] # for class and functions

    # Exclude class and function definitions from all_blocks
    for chunk in all_blocks:
        stripped_chunk = chunk.strip()
        if not (stripped_chunk.startswith("class ") or stripped_chunk.startswith("def ")):
            filtered_chunks.append(chunk)
        else:
            if block != chunk:
                other_chunks.append(chunk)

    all_words = set()

    # Manage imports
    for line in filtered_chunks:
        if line.startswith("import ") or line.startswith("from "):
            if "from " in line:
                package_name = line.split("import")[0].split("from")[1].strip()
            else:
                package_name = line.split("import")[1].strip()
                
            if len(package_name) > 0:
                # if package_name not start with .
                # go to our root folder and try and find the file
                # if package_name.startswith(".") or isFileInRoot(package_name) == True:
                char = extract_words_from_code(line.split("import")[1])
                for word in char:
                    if word in filtered_words:
                        all_words.add(word)
                        importLine.append({"packagePath":package_name,"imports":word})
        else:
            # variables line declaration
            # print("here with line ", line)
            varNames = get_assigned_variables(line)
            for word in varNames:
                if word in filtered_words:
                    all_words.add(word)
                    importLine.append({"packagePath":".","imports":word})

    # manage class and functions
    for chunk in other_chunks:
        # print(chunk)
        name = chunk.split("(")[0]
        name = name.replace("def","")
        name = name.replace("class","").strip()
        # print(name)
        if name in filtered_words:
            importLine.append({"packagePath":".","imports":name})

    # check for in file dependencies here
    words = set()
    for chunk in all_blocks:
        var_name = get_assigned_variables(chunk,True)
        if isinstance(var_name,str):
            # print(get_assigned_variables(chunk,True))
            words.add(var_name)

    ref = get_refrenced_variables_from_file(block,words)

    importLine.extend(ref)

    return importLine

def get_code_block_names(code,block_name):
    # print(f"code {block_name} ",code, " \n\n\n")
    try:
        tree = ast.parse(code.strip())

        for node in ast.walk(tree):
            if isinstance(node, ast.FunctionDef):  # Function names
                return node.name == block_name
            elif isinstance(node, ast.ClassDef):  # Class names
                return node.name == block_name
            elif isinstance(node, ast.Assign):  # Direct assignments (x = 5, x, y = 10, 20)
                for target in node.targets:
                    if isinstance(target, ast.Name):
                        if target.id == block_name: return True
    except:
        # pass
        # print(code," error jumbo ",block_name)

        return False

def getCodeBlockFromFile(blockName, file_dependencies):
    for chunk in file_dependencies:
        if get_code_block_names(chunk,blockName):
            return chunk
    return None

def removeDuplicates(code):
    """
    Removes duplicate code chunks from the given list, preserving the original order.

    Args:
        code (list): A list of strings, each representing a code chunk (can be multiline).

    Returns:
        list: A new list containing the unique code chunks from the input, in their original order.
    """
    cleaned_code = []
    for line in code:
        if line not in cleaned_code:
            cleaned_code.append(line)

    return cleaned_code

def get_if_blocks(code):
    tree = ast.parse(code.strip())
    if_blocks = []

    for node in ast.walk(tree):
        if isinstance(node, ast.If):  # Capture if statements
            if_block = {
                "type": "if",
                "condition": ast.unparse(node.test) if hasattr(ast, "unparse") else "<condition>",
                "body": ast.unparse(node.body) if hasattr(ast, "unparse") else "<body>",
            }
            if_blocks.append(if_block)

            # Process the `orelse` part which may contain elif or else blocks
            else_body = []
            for elif_node in node.orelse:
                if isinstance(elif_node, ast.If):  # Elif case
                    elif_block = {
                        "type": "elif",
                        "condition": ast.unparse(elif_node.test) if hasattr(ast, "unparse") else "<condition>",
                        "body": ast.unparse(elif_node.body) if hasattr(ast, "unparse") else "<body>",
                    }
                    if_blocks.append(elif_block)
                else:  # Else block (contains multiple statements)
                    else_body.append(ast.unparse(elif_node) if hasattr(ast, "unparse") else "<body>")

            if else_body:  # Capture the entire else block as a single entity
                else_block = {
                    "type": "else",
                    "body": "\n".join(else_body)
                }
                if_blocks.append(else_block)

    return if_blocks

def getCodeBlockNameAndType(code, returnVal=False):
    # print(code)
    tree = ast.parse(code.strip())

    for node in ast.walk(tree):
        if isinstance(node, ast.Assign):  # Direct assignments (x = 5, x, y = 10, 20)
            for target in node.targets:
    
                if isinstance(target, ast.Name):
                    var_name =  target.id
                
                     # Get the assigned value
                    value = node.value
                    var_type = None  # Default type

                    if isinstance(value, ast.Constant):  # Handles literals (Python 3.8+)
                        var_type = type(value.value).__name__
                    elif isinstance(value, ast.List):
                        var_type = "list"
                    elif isinstance(value, ast.Dict):
                        var_type = "dict"
                    elif isinstance(value, ast.Tuple):
                        var_type = "tuple"
                    elif isinstance(value, ast.Set):
                        var_type = "set"
                    elif isinstance(value, ast.Call):  # Function call
                        var_type = "function_call"
                    elif isinstance(value, ast.BinOp):  # Binary operations (e.g., x + y)
                        var_type = "expression"

                    # print("data ", var_name," ",var_type, " ", val_value)
                    if returnVal:
                        return [var_name,var_type,value]
                    
                    return [var_name,var_type]
                
        elif isinstance(node, ast.If):  # Capture if statements
            condition = ast.unparse(node.test) if hasattr(ast, "unparse") else "<condition>"
            # print(condition, " is this")
            return ["if_"+condition,"if_statement"]

    return None

def remove_multiline_comments(code: str) -> str:
    """
    Removes all multi-line comments (triple-quoted strings that are not docstrings) from Python code.
    """
    pattern = re.compile(r'(""".*?"""|\'\'\'.*?\'\'\')', re.DOTALL)
    
    def replacer(match):
        # Keep docstrings in functions and classes
        before = code[:match.start()].strip().split('\n')
        if before and (before[-1].startswith("def ") or before[-1].startswith("class ")):
            return match.group(0)
        return ""
    
    return pattern.sub(replacer, code)

def removeMethodsFromChunk(chunk):
    """
        Remove methods from class definitions
    """
    if not chunk.strip().startswith("class"):
        return chunk
    
    lines = chunk.split("\n")
    classWithoutMethod = [lines.pop(0)]

    for line in lines:
        stripped_line = line.strip()
        if stripped_line.startswith("def ") or stripped_line.startswith("@"):
            break
        classWithoutMethod.append(line)
    
    return "\n".join(classWithoutMethod)

def arrangeChunks(data,arranged_chunks,processed):
    code_to_name = {}
    sorted_chunks = []

    for chunk in data:
        name = get_assigned_variables(chunk,True)
        # print("name is ",name," with type ",type(name))
        if isinstance(name, str):
            # print("chunk name is ",name)
            code_to_name[name] = chunk

    chunk_order = getArrangeChunksOrder(data,arranged_chunks,processed)
    # print("chunk order is ", chunk_order)

    for i in chunk_order:
        sorted_chunks.append(code_to_name[i])

    return sorted_chunks

def getArrangeChunksOrder(data,arranged_chunks,processed):
    stack = []

    data = [removeMethodsFromChunk(i) for i in data]
    # print(data, "\n\ncleaned\n\n")

    for chunk in data:
        name = get_assigned_variables(chunk,True)
        # chunk_without_method = removeMethodsFromChunk(chunk)
        # print("name of chunk is ", chunk)
        dep = getBlockDependencies(chunk,data)
        dep = [i['imports'] for i in dep if i['packagePath'] == "."]

        if len(dep) == 0:
            arranged_chunks.append(chunk)
            processed.append(name)
        else:
            addToList = True
            for c in dep:
                if c not in processed:
                    addToList = False
                    stack.append(chunk)
                    break

            if addToList:
                arranged_chunks.append(chunk)
                processed.append(name)

    if len(stack) > 0:
        # print(stack)
        # print("\n\nprocessed\n",processed)
        arranged_chunks = getArrangeChunksOrder(stack,arranged_chunks,processed)

    return [name for name in processed if isinstance(name,str)] #arranged_chunks



import asyncio

from .src.Express.extract.reverseImport import convertToImportStatements
from .parsers.javascript_typescript.jsParser import JsTxParser


parser = JsTxParser()

#     print("Parsing Test.js")
imports, chunks, chunks_name, import_deps = asyncio.run(parser.parse_code("/home/attah/Documents/sb_final/speedbuild/test.js"))
print(import_deps)

o=convertToImportStatements(import_deps)

print(o)


import os
import json
import time
import shutil
import tempfile
import threading
import subprocess
import importlib.util
import concurrent.futures

from pathlib import Path
from importlib import metadata

from ....utils.pushPackage import pushPythonPackageToServer

CACHE_FILE = str(Path.home()) + "/.sb/sb_app_mapping_cache.json"
CACHE_LOCK = threading.Lock()  # For thread-safe cache operations
MAX_WORKERS = 5  # Reduced from 5 to avoid resource exhaustion
TIMEOUT = 120  # Seconds to wait before considering a package installation hung

def load_django_settings(settings_path):
    """Loads the Django settings module dynamically."""
    spec = importlib.util.spec_from_file_location("settings", settings_path)
    settings_module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(settings_module)
    return settings_module

def create_temp_env():
    """Creates a temporary virtual environment and returns its path."""
    temp_dir = tempfile.mkdtemp()
    venv_path = os.path.join(temp_dir, "venv")
    try:
        # Add timeout to avoid hanging
        result = subprocess.run(
            ["python", "-m", "venv", venv_path], 
            check=True, 
            timeout=30  # 30 seconds should be enough for venv creation
        )
        return temp_dir, venv_path
    except (subprocess.TimeoutExpired, subprocess.SubprocessError) as e:
        print(f"Error creating virtual environment: {str(e)}")
        shutil.rmtree(temp_dir)
        raise

def get_site_packages_path(venv_path):
    """Returns the site-packages path for the virtual environment."""
    python_bin = os.path.join(venv_path, "Scripts" if os.name == "nt" else "bin", "python")
    try:
        output = subprocess.run(
            [python_bin, "-c", "import site; print(site.getsitepackages()[0])"],
            capture_output=True,
            text=True,
            timeout=10  # 10 seconds timeout
        )
        return output.stdout.strip()
    except subprocess.TimeoutExpired:
        print(f"Timeout getting site-packages path for environment: {venv_path}")
        raise

def install_package(venv_path, package_name):
    """Installs a package in the temporary virtual environment."""
    pip_bin = os.path.join(venv_path, "Scripts" if os.name == "nt" else "bin", "pip")
    try:
        print(f"Starting installation of {package_name}...")
        process = subprocess.run(
            [pip_bin, "install", package_name], 
            check=True, 
            timeout=TIMEOUT,  # Timeout to prevent hanging
            capture_output=True,
            text=True
        )
        print(f"Successfully installed {package_name}")
        return True
    except subprocess.TimeoutExpired:
        print(f"Installation of {package_name} timed out after {TIMEOUT} seconds")
        return False
    except subprocess.SubprocessError as e:
        print(f"Error installing {package_name}: {str(e)}")
        return False

def detect_installed_apps(site_packages_path):
    """Detect likely Django apps in site-packages."""
    try:
        apps = []
        for item in os.listdir(site_packages_path):
            full_path = os.path.join(site_packages_path, item)

            # Ignore metadata folders
            if item.endswith(".dist-info") or item.endswith(".egg-info"):
                continue

            # Include real directories
            if os.path.isdir(full_path):
                # Check for __init__.py to confirm it's a module
                # if os.path.isfile(os.path.join(full_path, "__init__.py")):
                #     apps.append(item)
                apps.append(item)
        return apps
    except Exception as e:
        print(f"Error detecting installed apps: {str(e)}")
        return []

def safe_load_json(file_path):
    """Safely load JSON from a file with retries."""
    max_retries = 3
    for attempt in range(max_retries):
        try:
            with open(file_path, "r") as f:
                return json.load(f)
        except json.JSONDecodeError:
            if attempt < max_retries - 1:
                time.sleep(0.5)  # Wait before retrying
            else:
                return {}
        except FileNotFoundError:
            return {}

# here
def save_cache(data):
    """
    Saves the provided data to a cache file, merging it with any existing cached data.
    Args:
        data (dict): The data to be saved to the cache. This will be merged with the existing cache.
    Raises:
        Exception: If there is an error while writing to the cache file, an error message is printed.
    """
    """Saves the app mapping to a cache file."""
    with CACHE_LOCK:
        # First load existing cache to merge with new data
        existing_cache = safe_load_json(CACHE_FILE)

        if existing_cache == {}:
            os.makedirs(CACHE_FILE, exist_ok=True)

        # Only update if there is a difference
        if any(existing_cache.get(k) != v for k, v in data.items()):
            # existing_cache.update(data)
            # send data to server for update and pull all package data
            res = pushPythonPackageToServer(data)
            if res is not None:
                # set existing_cache to server response
                existing_cache = res
            else:
                # merge offline data
                # Merge and save
                existing_cache.update(data)

        try:
            with open(CACHE_FILE, "w") as f:
                json.dump(existing_cache, f, indent=4)
        except Exception as e:
            print(f"Error saving cache: {str(e)}")

def load_cache():
    """Loads the app mapping from cache if available."""
    with CACHE_LOCK:
        return safe_load_json(CACHE_FILE)

def process_single_package(package_name):
    """Process a single package in its own virtual environment."""
    print(f"Processing package: {package_name}")
    cache = load_cache()
    
    # Skip if already in cache
    if package_name in cache:
        print(f"Using cached data for {package_name}")
        return package_name, cache[package_name]
    
    temp_dir = None
    try:
        temp_dir, venv_path = create_temp_env()
        site_packages = get_site_packages_path(venv_path)
        
        success = install_package(venv_path, package_name)
        if not success:
            print(f"Installation failed for {package_name}")
            return package_name, []
            
        installed_apps = detect_installed_apps(site_packages)
        print(f"Found apps for {package_name}: {installed_apps}")
        
        # Update cache with this result
        cache_update = {package_name: installed_apps}
        save_cache(cache_update)
            
        return package_name, installed_apps
    except Exception as e:
        print(f"Error processing {package_name}: {str(e)}")
        return package_name, []
    finally:
        # Cleanup temp environment
        if temp_dir and os.path.exists(temp_dir):
            try:
                shutil.rmtree(temp_dir)
                print(f"Cleaned up environment for {package_name}")
            except Exception as e:
                print(f"Failed to clean up {temp_dir}: {str(e)}")


def get_package_version(package_name):
    try:
        return metadata.version(package_name)
    except metadata.PackageNotFoundError:
        return "unknown"

def get_django_app_from_packages_parallel(package_names, max_batch_size=10):
    """
    Process multiple package names in parallel to retrieve Django applications.
    This function processes a list of package names to identify Django applications within them,
    utilizing parallel processing for efficiency. It implements caching to avoid reprocessing
    previously analyzed packages.
    Args:
        package_names (list): A list of package names to process.
        max_batch_size (int, optional): Maximum number of packages to process in parallel. 
            Defaults to 10.
    Returns:
        dict: A dictionary mapping package names to lists of discovered Django applications.
            For packages that fail processing or timeout, an empty list is returned.
    Example:
        >>> packages = ['django-allauth', 'django-rest-framework']
        >>> results = get_django_app_from_packages_parallel(packages)
        >>> print(results)
        {'django-allauth': ['allauth', 'allauth.account'], 'django-rest-framework': ['rest_framework']}
    Note:
        - Uses threading for parallel processing
        - Implements timeout handling to prevent hanging
        - Caches results for future lookups
        - Processes packages in batches to manage resource usage
    """
    cache = load_cache()
    results = {}
    # packages_to_process = [pkg for pkg in package_names if pkg not in cache]
    packages_to_process = []
    
    # Add cached results directly
    for pkg in package_names:
        if pkg in cache:
            results[pkg] = {"pkg":cache[pkg],"version":get_package_version(pkg)}
            # print(get_package_version(pkg), pkg, "package version")
        else:
            packages_to_process.append(pkg)
        
    
    # Process in smaller batches to prevent resource exhaustion
    for i in range(0, len(packages_to_process), max_batch_size):
        batch = packages_to_process[i:i+max_batch_size]
        print(f"Processing batch of {len(batch)} packages")
        
        with concurrent.futures.ThreadPoolExecutor(max_workers=MAX_WORKERS) as executor:
            future_to_package = {
                executor.submit(process_single_package, pkg): pkg 
                for pkg in batch
            }
            
            for future in concurrent.futures.as_completed(future_to_package):
                try:
                    package_name, installed_apps = future.result(timeout=TIMEOUT + 30)
                    
                    results[package_name] = {"pkg":installed_apps,"version":get_package_version(pkg)}#installed_apps
                except concurrent.futures.TimeoutError:
                    package = future_to_package[future]
                    print(f"Processing timed out for package: {package}")
                    results[package] = {"pkg":[],"version":None}
                except Exception as e:
                    package = future_to_package[future]
                    print(f"Exception for package {package}: {str(e)}")
                    results[package] = {"pkg":[],"version":None}
    
    return results

def get_installed_packages(venv_path):
    """Get list of installed packages in the given virtual environment."""
    pip_path = os.path.join(venv_path, "bin", "pip")  # Linux/macOS

    if os.name == "nt":
        pip_path = os.path.join(venv_path, "Scripts", "pip.exe")  # Windows

    try:
        result = subprocess.run(
            [pip_path, "list"], 
            capture_output=True, 
            text=True,
            timeout=30  # 30 seconds timeout
        )
        lines = result.stdout.splitlines()[2:]  # Skip headers
        packages = {line.split()[0] for line in lines}  # Extract package names
        return packages
    except subprocess.TimeoutExpired:
        print("Timeout getting installed packages")
        return set()
    except Exception as e:
        print(f"Error getting installed packages: {str(e)}")
        return set()

def getDjangoAppsPackage(settings_path, venv):
    """
    Maps Django installed apps to their corresponding Python packages.
    This function analyzes the Django settings file to find installed apps and determines
    which Python packages provide those apps. It handles both direct app imports and
    regular Python package imports.
    Args:
        settings_path (str): Path to the Django settings.py file
        venv (str): Path to the Python virtual environment to analyze
    Returns:
        dict: A mapping of Django app names to lists of package names that provide them.
              Returns empty dict if settings file is not found or cannot be loaded.
              Format: {'app_name': ['package1', 'package2']}
    Raises:
        No exceptions are raised - errors are handled internally and logged to stdout
    Example:
        >>> getDjangoAppsPackage('/path/to/settings.py', '/path/to/venv')
        {'django.contrib.admin': ['django'],
         'django.contrib.auth': ['django'],
         'myapp': ['my-package']}
    """
    if not os.path.exists(settings_path):
        print(f"Error: Settings file '{settings_path}' not found!")
        return {}

    # Load Django settings
    try:
        settings_module = load_django_settings(settings_path)
        installed_apps = getattr(settings_module, "INSTALLED_APPS", [])
    except Exception as e:
        print(f"Error loading Django settings: {str(e)}")
        return {}

    # Get all installed packages
    packages = get_installed_packages(venv)
    
    # Process packages in parallel
    package_app_mapping = get_django_app_from_packages_parallel(packages)
    
    # Map Django apps to their packages
    app_mapping = {}
    for package, django_apps in package_app_mapping.items():
        for app in django_apps['pkg']:
            if app in installed_apps:  # or is a regular import
                if app in app_mapping:
                    app_mapping[app]['pkg'].append(package)
                else:
                    app_mapping[app] = {"pkg":[package]}

                    # check if version is not None or Unkmown
                    pkg_version = django_apps['version'] 
                    if pkg_version != None or pkg_version != "unknown":
                        app_mapping[app]['version'] = pkg_version 

    return app_mapping

# Example usage
# if __name__ == "__main__":
#     # For testing
#     settings_path = "path/to/your/settings.py"
#     venv_path = "path/to/your/venv"
    
    # Uncomment to run
    # app_mapping = getDjangoAppsPackage(settings_path, venv_path)
    # print(json.dumps(app_mapping, indent=2))


"""
Deploy, Test and Customize Express SpeedBuild Features
"""
import os
import time
from pathlib import Path
from rich.prompt import Prompt

from ....break_prompt import splitPrompt
from ....cli_output.output import StatusManager
from .moveToTarget import moveTemplateFilesToTargetProject
from .debug import debug_and_customize_template_code,debugTemplate,test_environment_path

home = str(Path.home())
async def deployExpressFeature(template_name,deploy_project_root,customize_prompt=False):
    """
    Deploys an Express.js feature template to a specified project root directory.
    This function handles the deployment process of Express.js feature templates, including
    testing, and optional customization of the template code.

    Args:
        template_name (str): Name of the Express.js template to deploy
        deploy_project_root (str): Root directory path where the template will be deployed
        customize (bool, optional): Flag to enable template customization. Defaults to False

    Raises:
        ValueError: If the specified template does not exist in the templates directory

    Example:
        deployExpressFeature("auth-template", "/path/to/project", customize=True)

    Note:
        - The function creates a test environment if it doesn't exist
        - Templates are tested before deployment using 'npm run dev'
        - If customization is enabled, user will be prompted for customization details
    """

    logger = StatusManager()

    logger.print_message("Starting Feature deployment")

    templates_path = f"{home}/.sb_zip"

    # create test environment if it does not already exist.
    if not os.path.exists(test_environment_path):
        os.makedirs(test_environment_path, exist_ok=True)

    # get template
    # done : remember to update this when u've implemented template versioning for express
    template_version = "lts"
    if ":" in template_name:
        template_name, template_version = template_name.split(":")

    # check if template exists
    if not os.path.exists(f"{templates_path}/{template_name}/{template_version}.zip"):
        raise ValueError(f"Template with the name {template_name} and version {template_version} does not exist")

    path_to_template = f"{templates_path}/{template_name}/{template_version}.zip"
    
    debugTemplate(path_to_template)

    # check for and perform customization
    if customize_prompt:
        prompt = Prompt.ask("[cyan]How do you want to customize feature code (be as detailed as possible)[/cyan]")
        logger.print_message("Customizing code, please wait")

        # done : break customization prompt into sub prompts
        customize_prompts = splitPrompt(prompt)
        for prompt in customize_prompts:
            debug_and_customize_template_code(None,prompt)
            time.sleep(0.5)

        logger.print_message("Checking for errors")
        # debug customization logic
        debug_and_customize_template_code("npm run dev")

    logger.stop_status()
    # done : transplant template code into target environment
    await moveTemplateFilesToTargetProject(deploy_project_root)
    logger.print_message("Feature deployment complete")



import os
import json
import shutil
import tempfile
from ..utils.state import getProjectPathFromId


def swap_folder_contents(folder_a, folder_b):
    if not os.path.isdir(folder_a) or not os.path.isdir(folder_b):
        raise ValueError("Both paths must be existing folders.")

    # Create a temporary directory
    with tempfile.TemporaryDirectory() as temp_dir:
        # Move contents of A -> temp
        for item in os.listdir(folder_a):
            shutil.move(os.path.join(folder_a, item), temp_dir)

        # Move contents of B -> A
        for item in os.listdir(folder_b):
            shutil.move(os.path.join(folder_b, item), folder_a)

        # Move contents of temp -> B
        for item in os.listdir(temp_dir):
            shutil.move(os.path.join(temp_dir, item), folder_b)


def updateStateDotJSON(path,actions,undo):
    if os.path.exists(path):
        with open(path,"w") as file:
            json.dump({
            "action": actions,
            "undo": not undo
        }, file, indent=4)
    else:
        raise ValueError(f"Something went wrong, cannot find {path}")


def getFileStateAndCurrentPath(fileName,project_state_path,current_project_path):
    current_file_path = os.path.join(current_project_path, fileName.lstrip("/")) #this take file name plus folder name

    # fileName = fileName.lstrip("/").split("/")[-1] #get only filename, cause we dont store parent folders in state
    last_sb_state_file_path = os.path.join(project_state_path, "last_sb_update", fileName)
    prev_state_file_path = os.path.join(project_state_path, "prev", fileName)

    return [prev_state_file_path,last_sb_state_file_path,current_file_path]
                    

def FileIsSameAsPrevState(prev_file_state, last_update_file):
    """
    Compare the contents of file_name in the current project path and in the 'prev' directory.
    Returns True if the files are exactly the same, False otherwise.
    """

    if not os.path.exists(prev_file_state) or not os.path.exists(last_update_file):
        return False

    with open(prev_file_state, "rb") as f1, open(last_update_file, "rb") as f2:
        return f1.read() == f2.read()
    
def getTemplateFileNames(path):
    file_paths = []
    for root, _, files in os.walk(path):
        if "node_modules" in root:
            continue
        for file in files:
            # Get the absolute path by joining root and file
            absolute_path = os.path.join(root, file)
            file_paths.append(absolute_path.replace(path,"").lstrip("/"))
    return file_paths


def undoSBDeploy():

    current_project_path = os.getcwd()

    config_path = os.path.join(current_project_path,".speedbuild")

    if not os.path.exists(config_path):
        print("Undo was Unsuccessful")
        return
    
    with open(config_path,"r") as sbFile:
        project_id = sbFile.read()
    
    # get project path
    project_state_path = getProjectPathFromId(project_id)
    undo_actions = {}

    state_file_path = f"{project_state_path}/state.json"
    state_prev_folder = os.path.join(project_state_path, "prev")
    state_last_sb_update_folder = os.path.join(project_state_path, "last_sb_update")

    if os.path.exists(state_file_path):
        with open(state_file_path,"r") as file:
            data = json.loads(file.read())

            actions = data['action']
            undo = data['undo']

            # If one file in prev state is not matching current project state, Abort undo
            files_in_prev = getTemplateFileNames(state_prev_folder)#os.listdir(state_prev_folder)
            proceed_with_undo = True

            for step in actions.keys():
                # fileName = step.lstrip().split("/")[-1]
                if step in files_in_prev:
                    # check if its same
                    prev_state_file_path,last_sb_state_file_path,current_file_path = getFileStateAndCurrentPath(step,project_state_path,current_project_path)
 
                    if os.path.exists(current_file_path):
                        if not FileIsSameAsPrevState(current_file_path,last_sb_state_file_path):
                            proceed_with_undo = False

            if not proceed_with_undo:
                print("Files has changed too much to proceed with undo")
                return
            
            for step in actions.keys():
                last_file_action = actions[step]

                prev_state_file_path,last_sb_state_file_path,current_file_path = getFileStateAndCurrentPath(step,project_state_path,current_project_path)

                if last_file_action == "update" or last_file_action == "delete":
                    with open(current_file_path,"w") as file:
                        with open(prev_state_file_path,"r") as prevFile:
                            file.write(prevFile.read())

                        # update state action
                        undo_actions[step] = "update" if last_file_action=="update" else "create"

                elif last_file_action == "create":
                    #delete file
                    if os.path.exists(current_file_path):
                        os.remove(current_file_path)

                    # update state action
                    undo_actions[step] = "delete"

    # swap prev and last_update folder
    swap_folder_contents(state_prev_folder,state_last_sb_update_folder)

    # update state.json
    updateStateDotJSON(state_file_path,undo_actions,data['undo']) 

    print("finished here")


import asyncio
from .jsParser import JsTxParser


def parse_object_exports(exports_str):
    """Parse object exports respecting nested braces"""
    result = []
    current = ""
    depth = 0
    
    for char in exports_str:
        if char == '{':
            depth += 1
        elif char == '}':
            depth -= 1
        elif char == ',' and depth == 0:
            if current.strip():
                result.append(current.strip())
            current = ""
            continue
        current += char
    
    if current.strip():
        result.append(current.strip())
    
    return result

async def getModuleExport(filename,deps_set):
    """
        Get export statements of dependencies in a CommonJS/ES6 module.

        Args:
            filename : Path to JS module file
            deps_set : set of dependency names to get export statements for

        Returns:
            list of export statement strings
    """

    parser = JsTxParser() # js parser
    _, chunks, _, _ = await parser.parse_code(filename) #parse file code

    module_export = []

    for chunk in chunks:
        if chunk.strip().startswith("module.exports") or chunk.strip().startswith("export default "):

            if "=" not in chunk:  # Skip if no assignment
                default_es6_export = chunk.replace("export default ","").strip()
                if default_es6_export in deps_set:
                    module_export.append(chunk)

                continue

            declaration,exports = chunk.split("=",1)
            declaration = declaration.strip()

            if declaration == "export default":
                continue

            # process declaration
            if declaration != "module.exports":
                declaration = declaration.replace("module.exports.","").strip()
                if declaration in deps_set:
                    module_export.append(chunk)
                    continue

            # process exports
            exports = exports.strip().rstrip(";")
            if exports.startswith("{") and exports.endswith("}"):
                exports = exports.lstrip("{").rstrip("}")

                # Use smart parser that respects nesting
                export_items = parse_object_exports(exports)

                # module.exports = {foo, bar: {name: "james"}, baz}
                export_lists = []
                for item in export_items:
                    if ":" in item: # meaning we are exporting an object
                        key = item.split(":", 1)[0].strip() # get the key, value pair
                        if key in deps_set:
                            export_lists.append(item)
                    elif item.strip() in deps_set: # we are exporting named exports
                        export_lists.append(item.strip()) # get the named export

                if export_lists:
                    export_data = f"{declaration} = {{ {', '.join(export_lists)} }}"
                    module_export.append(export_data)

            elif exports in deps_set:
                # e.g module.exports = number;
                module_export.append(chunk)

    return module_export


# res = getModuleExport("/home/attah/Documents/speedbuildjs/express.js_contact_app/controllers/contact_controller.js",["getContacts","createContact","first"])
# print(res)


import time
import requests

from .cli_utils import get_sb_config,updateAuthToken
from .ws_utils import caculateTimeDifference


max_access_lifespan = 5
base_url = "https://backend.speedbuild.dev/"


def askUserToLogin():
    print("Your session has expired\nYou need to login")


def getNewAccessToken(refreshToken):
    data = {"refresh":refreshToken}

    isValid, resData = manageServerCall('POST','api/token/refresh/',data)
    
    if isValid:
        return resData['access']
    
    # refesh token has expired
    # prompt the user to login
    return askUserToLogin()

def getRequestHeaders():
    user_data = get_sb_config()
    if "jwt_tokens" not in user_data:
        return None
    
    user_data = user_data['jwt_tokens']

    # check if user_data has access and refresh token
    if 'access_token' not in user_data.keys() or 'refresh_token' not in user_data.keys():
        return None
    
    diff = caculateTimeDifference(user_data['time_created'])
    
    if diff > max_access_lifespan:
        # access token has expired
        # request new access token with current refresh token
        refresh_token = user_data['refresh_token']
        access_token = getNewAccessToken(refresh_token)

        if access_token != None:
            # refresh and access token is valid
            # save new access token to json file
            # user_data['access_token'] = access_token
            # user_data["time_created"] = time.time()
            # saveUserDataToLocalStorage(user_data)
            updateAuthToken({
                "access_token" :access_token,
                "refresh_token":refresh_token,
                "time_created":time.time()
            })
        else:
            # refresh token has expired
            print("Session has expired please login")
    else:
        access_token = user_data['access_token']

    headers = {
        "Authorization" : f"Bearer {access_token}",
    }

    return headers


def manageServerCall(method,path,data={},files={},useAuthentication=False):
    url = base_url + path
    headers = {} #{"Content-Type": "application/json"}

    if useAuthentication:
        headers = getRequestHeaders()

        if headers == None:
            print("You need to login")
            return 0


    if method == "GET":
        res = requests.get(url,headers=headers)
    elif method == "POST":
        res = requests.post(url,data=data,files=files,headers=headers)
    elif method == "DELETE":
        res = requests.delete(url,headers=headers)


    if res.ok:
        return [True,res.json()]
    
    elif 'detail' in res.json().keys():
        return [False,None]

    raise Exception(f"Server responded with status code {res.status_code} and error {res.reason}")


def pingServer():
    valid,res = manageServerCall('POST','sb/log-action/',{'command':'ping'},{},True)
    if valid:
        print(res)

def logUserAction(command):
    manageServerCall('POST','sb/log-action/',{'command':command},{},True)


from typing import Optional
from pydantic import BaseModel, Field

class LLMOutput(BaseModel):
    """
    Represents the output of a language model (LLM) action.

    Attributes:
        action (str): Action to perform.
        write_action (Optional[str]): Specifies the code write action to perform (add, replace, or remove).
        line (Optional[str]): Line(s) of code to replace or remove.
        code (str): New code to write to the file.
        command (Optional[str]): CLI command to execute.
        file_path (Optional[str]): Path of the file to read.
    """
    action: str = Field(description="Action to perform")
    write_action: Optional[str] = Field(
        default=None, 
        description="What code write action to perform (add, replace or remove)"
    )
    line: Optional[str] = Field(
        default=None, 
        description="line(s) of code to replace or remove"
    )
    code: str = Field(
        default=None, 
        description="New code to write to file"
    )
    command: Optional[str] = Field(
        default=None,
        description="CLI command to execute"
    )
    file_path: Optional[str] = Field(
        default=None,
        description="path of file to read"
    )


system_prompt = """
You are an expert Django developer debugging a Django project. 
You are given error logs, stack traces, or failing tests and must identify and fix bugs.

## **Rules**

* Perform exactly **ONE action per turn**
* Never touch `requirements.txt` - ignore it completely
* Make minimal changes only
* Explain your reasoning before each action

## **Available Tools**

* `readFile` - Read a file
      Return the content of a file

      NOTE : always use the absolute path of the file !!!

      Args:
          filePath: first str
          fileName: second str

* `updateFile` - Update a file 
    Updates a file at the specified path by performing the given action.

    Args:
        filePath (str): The directory path where the file is located.
        fileName (str): The name of the file to update.
        action (str): The action to perform on the file. Accepted values are "add", "replace", "remove", or "create".
        update (str, optional): The content to add or replace in the file. Required for "add", "replace", and "create" actions.
        line (str, optional): The specific line number (not the actual line content) to update. Used for "replace" or "remove" actions.
    
    Example:
        //for single line action
        updateFile('./output','hello.js','replace',"console.log('hello')",'10')

        //for multi line action
        updateFile(
            './output',
            'hello.js',
            'replace',
            "console.log('hello'); console.log("world")",
            '10-12')
    
    Returns:
        str: An error message if the file is not found or the action is not recognized; otherwise, None.
    
    Notes:
        - For "add", "replace", and "remove" actions, the file must already exist.
        - For "create" action, a new file is created with the provided content.
        - The function relies on a helper function `performUpdate` to process the file content.


* `executeCommand` - Run a shell command
    Executes a given shell command using the PythonExecutor.

    Args:
        command (str): The shell command to execute as a string.
        framework (str) : django

    Behavior:
        - Splits the command string into a list of arguments.
        - Determines if the command should be self-exiting based on whether the last argument is "runserver".
        - Executes the command in a specified working directory and environment.
        - Return the standard output and standard error of the executed command.

## **Process**

1. Choose ONE tool based on the error
2. Explain why you're using this tool
3. Use the tool
4. Wait for next turn

Stop when the issue is resolved or you need human help.

"""


import subprocess
import sys
import os

def install_dependencies(dependencies, venv_path=None):
    """Install dependencies inside a virtual environment if provided, otherwise install globally."""
    
    # Determine the correct pip executable
    if venv_path:
        pip_executable = os.path.join(venv_path, "bin", "pip") if os.name != "nt" else os.path.join(venv_path, "Scripts", "pip.exe")
        if not os.path.exists(pip_executable):
            print(f"Error: Virtual environment not found at {venv_path}")
            return
    else:
        pip_executable = sys.executable.replace("python", "pip")  # Use the system's pip

    # Install dependencies
    for package in dependencies:
        try:
            subprocess.run([pip_executable, "install", package], check=True)
            print(f"Successfully installed {package}")
        except subprocess.CalledProcessError:
            print(f"Failed to install {package}")




import requests

from .server_address import url_address

def pushPythonPackageToServer(cache):
    url = url_address + "sb/packages/"
    headers = {"Content-Type": "application/json"}
    try:
        response = requests.post(
            url,
            json={"packages": cache},
            headers=headers,
            timeout=10
        )
        response.raise_for_status()
        return response.json() if response.status_code == 200 else None
    except (requests.RequestException, Exception):
        return None



import re

varIdentifies = ["const","let","var"]

# TODO : handle single and multi line comment
chunk = """const getContacts= asyncHandler(async (req,res)=>{
    const hello = 1
    const contacts = await Contact.find({user_id:req.user._id})
    const name = hello
    res.status(200).json(contacts)
})"""

def removeCommentFromChunk(chunk):
    return chunk

# TODO : also collect variable name
def getLineValue(line):
    if "=" in line:
        pos = line.index("=") + 1
        return [line[:pos],line[pos:]]
    
    return None

# def getWordsInLine(line):
#     pass

def getWordsInLine(code, removeStrings=True):
    if code == None:
        return set()
    # Remove strings and comments
    if removeStrings:
        code = re.sub(r'(".*?"|".*?")', '', code)  # Remove strings
        
    code = re.sub(r'#.*', '', code)  # Remove comments
    
    # Extract words using regex (identifiers, keywords, function names, variable names)
    words = re.findall(r'[a-zA-Z_][a-zA-Z0-9_]*', code)
    
    return set(words)

def getChunkDependencies(chunk,chunks_name,raw_code=True,extra_deps=None):

    if raw_code: # TODO : investigate this 
        if chunk not in chunks_name.keys():
            print("getting ",chunk)
            print(chunks_name)
            raise ValueError("The code you want to parse is not in the file")
        
        chunk = chunks_name[chunk]

    deps = set(chunks_name)
    
    if extra_deps is not None:
        deps = deps.union(set(extra_deps))

    all_deps = set()
    exclude_deps = set()

    # remove comments from chunks
    chunk = removeCommentFromChunk(chunk)
    
    lines = chunk.split("\n")
    for line in lines:
        line = line.strip()
        line_word = line.split(" ")

        if line_word[0] in varIdentifies:

            line_diff = getLineValue(line)

            if line_diff is not None:
                # print(line_diff)
                beforeEqualSign,afterEqualSign = line_diff

                words_in_line = getWordsInLine(afterEqualSign)
                exclude_word = getWordsInLine(beforeEqualSign)

                chunk_dep = deps.intersection(words_in_line)
                
                exclude_deps= exclude_deps.union(exclude_word)
                # print(chunk_dep)
                all_deps = all_deps.union(chunk_dep)
        else:
            words_in_line = getWordsInLine(line)
            chunk_dep = deps.intersection(words_in_line)
            all_deps = all_deps.union(chunk_dep)


    if len(all_deps) > 0:
        # print(all_deps, " ", exclude_deps, " ", all_deps.difference(exclude_deps))
        return all_deps.difference(exclude_deps)
        
    return []



import os
import shutil

from .django_app_dependencies import get_package_version, load_cache

from ....utils.utils import findFilePath
from .django_utils import django_defaults
from ....parsers.python.parser import PythonBlockParser
from .write_dependency import OneStep, writeCodeToFile
from ....utils.package_utils import get_package_name, is_standard_library
from .feature_dependencies import extract_words_from_code, get_code_block_names, getBlockDependencies, getFileDependencies


def clear_folder(folder_path):
    for file_name in os.listdir(folder_path):
        file_path = os.path.join(folder_path, file_name)
        if os.path.isfile(file_path) or os.path.islink(file_path):  
            os.remove(file_path)  # Delete files & symlinks
        elif os.path.isdir(file_path):  
            shutil.rmtree(file_path)  # Delete directories
            
def getUrlPathForViews(feature,path):
    path_import = ["from django.urls import path"]
    urlpatterns = []

    with open(path,"r") as file:
            data = file.read()
            file_dependencies = PythonBlockParser().parse_code(data)
            chunkImport = [i for i in file_dependencies if i.startswith("import ") or i.startswith("from ")]

            file_dependencies = [chunk for chunk in file_dependencies if not (chunk.startswith("import ") or chunk.startswith("from "))]

            urlPaths = None

            for deb in file_dependencies:
                if get_code_block_names(deb,"urlpatterns"):
                    urlPaths = deb

            if urlPaths != None:
                urlPaths = urlPaths.split("\n")
                if len(urlPaths) > 1:
                    for path in urlPaths:
                        if feature in path:
                            if not path.endswith(","):
                                path += ","

                            urlpatterns.append(path)
                            view = path.split(",")[1].strip()

                            for importLine in chunkImport:
                                imports = importLine.split("import")[1]
                                imports = imports.replace("(","")
                                imports = imports.replace(")","").strip()

                                if "," in imports:
                                    imports = imports.split(",")
                                else:
                                    imports = [imports]

                                for i in imports:
                                    i = i.strip()
                                    if view.startswith(i):
                                        path_import.append(importLine.strip())
    return [urlpatterns, path_import]

def handleRouterURL(view_name,path):
    url_imports = ['from django.urls import path']
    url_content = []

    with open(path,"r") as file:
        data = file.read()
        chunks = PythonBlockParser().parse_code(data)
        feature = None
        initialized_pattern = False

        for chunk in chunks:
            if view_name in chunk and not(chunk.startswith("import ") or chunk.startswith("from ")):
                feature = chunk
                break
            elif "views" in chunk and (chunk.startswith("import ") or chunk.startswith("from ")):
                url_imports.append(chunk)

        router_var = feature.split(".")[0]
        router_words = [view_name]

        for i in chunks:
            if i.split("=")[0].strip() == router_var:
                var, val = i.split("=")
                if "(" in val:
                    val = val[:val.index("(")]
                router_words.extend([var.strip(),val.strip()])
                break

        for chunk in chunks:
            words_in_chunk = extract_words_from_code(chunk)
            intercept = words_in_chunk.intersection(router_words)
            
            if len(intercept) > 0:
                if chunk.startswith("import ") or chunk.startswith("from "):
                    url_imports.append(chunk)
                else:
                    if "urlpatterns" in chunk:
                        urlpatterns_var = chunk.split("=")[0].strip()

                        if initialized_pattern == False and urlpatterns_var != "urlpatterns":
                            url_content.append("urlpatterns = []")
                        else:
                            initialized_pattern = True
                            

                    url_content.append(chunk)

        # merge import and content
        url_imports = "\n".join(url_imports)
        url_content = "\n".join(url_content)

        url_code = f"{url_imports}\n\n{url_content}"

        # print(url_code)
        return [url_content,url_imports]


def getTemplateFileNames(path):
    file_paths = []
    for root, _, files in os.walk(path):
        for file in files:
            # Get the relative path to the base directory
            relative_path = os.path.relpath(os.path.join(root, file), path)
            file_paths.append(relative_path)
    return file_paths


def getURLForFeature(feature,file_path,app_name,project_path,output_dir):
    """
        if pattern not in urlpattern,
        check if pattern is in other chunks
        then get chunk dependencies and imports
    """

    path = file_path.split("/")
    path.pop()  #remove current file in path
    path.append("urls.py") #add urls.py to path
    path_list = path
    path = "/".join(path)
    parser = PythonBlockParser()

    template_file_path = "sb_app/urls.py"

    if os.path.exists(path):
        urlpatterns, imports = getUrlPathForViews(feature,path)

        if len(urlpatterns) > 0:
            urlpatterns = "\n".join(urlpatterns)
            urlpatterns = f"urlpatterns = [\n{urlpatterns}\n]"
            # imports = "\n".join(imports)
        else:
            urlpatterns, imports = handleRouterURL(feature,path)

        # print("url ",template_file_path, " ", output_dir)
        writeCodeToFile(template_file_path,urlpatterns,imports,False,output_dir)
    else:
        # TODO: get and loop through all url files in the project
        project_urls_files = findFilePath(project_path,"url]s.py")

        for url in project_urls_files:
            path = f"{project_path}/{url}"

            with open(path,"r") as file:
                data = file.read()
                chunks = parser.parse_code(data)
                for chunk in chunks:
                    if chunk.startswith(f"from {app_name}"):
                        urlpatterns, imports = getUrlPathForViews(feature,path)
   
                        if len(urlpatterns) > 0:
                            urlpatterns = "\n".join(urlpatterns)
                            urlpatterns = f"urlpatterns = [\n{urlpatterns}\n]"
                            # imports = "\n".join(imports)
                            # print("url ",template_file_path, " ", output_dir)
                            writeCodeToFile(template_file_path,urlpatterns,imports,False,output_dir)
                        break

def extract_feature_code_and_dependencies(
        feature_file_path,
        feature_name,
        project_path,
        project_dir,
        app_name,
        template_django_apps,
        settings,
        project_name,
        installed_apps,
        template_dep,
        processed,
        template_settings_import=[],
        template_confs = None,
        output_dir = None
    ):
    
    # print("feature file path ", feature_file_path)
    # print("output dir is ", output_dir)

    cache = load_cache()
    
    # get template conf from OneStep
    packages,use_email,settings_conf = OneStep(
        file_path=feature_file_path,
        feature=feature_name,
        project_root=project_path,
        folders_in_project_root=project_dir,
        writeToFile=True,
        app_name=app_name,
        project_name=project_name,
        use_email=False,
        settings_conf=[],
        output_dir = output_dir
    )

    if template_confs == None:
        template_confs =[]

    if len(packages) > 0:

        for package in packages:
            package = package.split(" ")[0].strip()

            if package in template_django_apps.keys():

                for pack in template_django_apps[package]['pkg']:
                    version = get_package_version(pack)
                    if version == "unknown":
                        version = None

                    if version:
                        template_dep.add(f"{pack}=={version}")
                    else:
                        template_dep.add(f"{pack}")
                installed_apps.add(package)
            else:

                probable_package = []

                if package not in cache.keys():
                    # TODO : work here ask the user for clarification

                    for pkg in cache:
                        if package in cache[pkg]:
                            probable_package.append(pkg) #Investigate here TODO

                    if len(probable_package) == 1:
                        package = probable_package[0]
                    elif len(probable_package) > 1:
                        # Ask user to select package that they used
                        while True:
                            user_input = input(f"Please which of this packages does ur feature depend on \n {probable_package}")
                            if user_input.strip() in probable_package:
                                package = user_input.strip()
                                break

                            print("Please Enter the correct package to proceed")
                    # print("couldnt find package ",package)
                    
                if len(probable_package) > 0:
                    version = get_package_version(package)
                    if version == "unknown":
                        version = None
                    if version:
                        template_dep.add(f"{package}=={version}")
                    else:
                        template_dep.add(package)


        conf_list = settings.getDjangoAppsConfigurations(list(installed_apps))

        conf_list.extend(settings_conf)

        # print("conf list = ",conf_list)

        remaining_conf = set(settings.custom_conf).difference(set(conf_list))

        if use_email and "EMAIL_BACKEND" in remaining_conf:
            conf_list.append("EMAIL_BACKEND")
            if "DEFAULT_FROM_EMAIL" in remaining_conf:
                conf_list.append("DEFAULT_FROM_EMAIL")


        conf_list = [i for i in conf_list if i not in processed][::-1]
        conf_queue = conf_list

        # get internal conf depenencies
        while len(conf_queue) > 0:
            conf = conf_queue.pop(0)

            if conf in processed or conf in django_defaults:
                # processed.add(conf)
                continue

            processed.add(conf)

            conf_code = settings.code_mappings[conf]

            # Here
            if conf_code not in template_confs:
                template_confs.append(conf_code) #add conf to list

            extFileDependencies = getFileDependencies(conf_code,project_name) # get external dependencies
            fileDependencies = getBlockDependencies(conf_code,settings.blocks) # in file dependencies

            # print("external conf dependencies ",extFileDependencies)
            # print("internal conf dependencies ", fileDependencies)

            # manage conf internal dependencies
            for dep in fileDependencies:
                source = dep['packagePath'].strip()
                imports = dep['imports'].strip()

                if source == imports:
                    template_settings_import.append(f"import {imports}")
                elif source != ".":
                    template_settings_import.append(f"from {source} import {imports}")
                else:
                    
                    if imports not in conf_list and imports not in processed:
                        conf_queue.append(imports)

            # manage conf external dependencies
            if len(extFileDependencies) > 0:
                for dep in extFileDependencies:
                    path, feature = dep
                    # print("handling dep ",dep)
                    main_dir = path.split("/")[0].strip()

                    # print(main_dir, " ",project_dir)

                    # if main_dir not in installed_apps:
                    # if main_dir in project_dir:
                    path = os.path.join(project_path,path)
                    if os.path.exists(path):

                        # TODO : sort conf here
                        # pop template_confs
                        # replace path
                        # re add to template_confs
                        code = template_confs.pop()

                        path = path.replace(".py","").replace(project_path,"")
                        if path.startswith("/"):
                            path = path[1:]

                        path_list = path.split("/")
                        start = path_list.pop(0)

                        if start == project_name:
                            path_list.insert(0,"<sb_ext>")
                        elif start == app_name:
                            path_list.insert(0,"<sb_ext_app>")
                        else:
                            path_list.insert(0,"<sb_ext_app>.sb_utils")

                        new_path = ".".join(path_list)

                        path = path.replace("/",".")
                        

                        code = code.replace(path,new_path)

                        # print("path to get external conf ",path, " feture ",feature)  

                        path = project_path + "/" + path.replace(".","/") + ".py"

                        # re add template conf
                        if code not in template_confs:
                            template_confs.append(code)


                        # Test environment start
                        _, template_confs, template_dep = extract_feature_code_and_dependencies(
                            path,
                            feature,
                            project_path,
                            project_dir,
                            app_name,
                            template_django_apps,
                            settings,
                            project_name,
                            installed_apps,
                            template_dep,
                            processed,
                            template_settings_import,
                            template_confs,
                            output_dir
                        )

                        # Test environment End
                                
                    else:
                        # print("here ok processinf conf for ",main_dir)
                        for item in cache:
                            if main_dir in cache[item]:
                                version = get_package_version(item)
                                if version == None or version == "unknown":
                                    template_dep.add(item)
                                else:
                                    template_dep.add(f"{item}=={version}")


                                # add main_dir to installed_apps so we can get configurations associated with package
                                installed_apps.add(main_dir)

                                new_conf = settings.getDjangoAppsConfigurations(list(installed_apps))
                                # print("\nfound new ",new_conf,"\n")

                                # check if main_dir is in source project installed apps
                                in_source_project_apps = settings.allSourceProjectApps
                                if main_dir not in in_source_project_apps:
                                    installed_apps.remove(main_dir)

                                # add to queue
                                for _conf in new_conf:
                                    if _conf not in conf_list and _conf not in processed: 
                                        conf_queue.append(_conf)

                                remaining_conf = remaining_conf.difference(conf_list)

                        # with open("sb_app_mapping_cache.json","r") as cacheFile:
                        #     cache = json.loads(cacheFile.read())
                        #     for item in cache:
                        #         if main_dir in cache[item]:
                        #             template_dep.add(item)

                        #             # add main_dir to installed_apps so we can get configurations associated with package
                        #             installed_apps.add(main_dir)

                        #             new_conf = settings.getDjangoAppsConfigurations(list(installed_apps))
                        #             # print("\nfound new ",new_conf,"\n")

                        #             # check if main_dir is in source project installed apps
                        #             in_source_project_apps = settings.allSourceProjectApps
                        #             if main_dir not in in_source_project_apps:
                        #                 installed_apps.remove(main_dir)

                        #             # add to queue
                        #             for _conf in new_conf:
                        #                 if _conf not in conf_list and _conf not in processed: 
                        #                     conf_queue.append(_conf)

                        #             remaining_conf = remaining_conf.difference(conf_list)

    return [template_settings_import, template_confs, template_dep]


import os
import re

from .manage_url import merge_urls
from .var_utils import get_assigned_variables
from ....parsers.python.parser import PythonBlockParser
from .feature_dependencies import arrangeChunks, getBlockDependencies, getCodeBlockFromFile, removeDuplicates


parser = PythonBlockParser()

def getWritePath(file_path, output_dir):
    path = file_path.split("/")
    write_dir = "output/"+output_dir+"/"

    if len(path) == 1 :
        dest = write_dir+path[0]
    
    else:
        
        file_name = path.pop()
        file_folder = path.pop()

        dest = f"{write_dir}{file_folder}"

        if not os.path.exists(dest):
            os.makedirs(dest)

        dest = f"{dest}/{file_name}"
    return dest


def writeCodeToFile(file_path,code,imports, writeAfter=False, output_dir=None):

    """
    write feature code to template file
    """

    dest = getWritePath(file_path,output_dir)

    if os.path.exists(dest):

        old_code = []
        old_imports = []

        # append to file
        with open(dest, "r+") as file:
            data = file.read()  # Step 1: Read existing content
        
            file.seek(0)        # Move to the beginning of the file
            file.truncate(0)    # Step 2: Clear the file

            old_content = parser.parse_code(data)
            for chunk in old_content:
                if chunk.startswith("import ") or chunk.startswith("from "):
                    # imports.extend([chunk])
                    old_imports.append(chunk)
                else:
                    old_code.append(chunk)

            # if writeAfter:
            #     write_data = f"{imports}\n{data}\n{code}"
            # else:
            #     write_data = f"{imports}\n{code}\n{data}"

            imports.extend(old_imports)
            old_code.insert(0,code)

            code = old_code

            # chunks = parser.parse_code(write_data)#split_code_into_sections(write_data)
            
            # for chunk in chunks:
            #     if chunk.startswith("import ") or chunk.startswith("from "):
            #         # imports.extend([chunk])
            #         imports.append(chunk)
            #     else:
            #         code.append(chunk)

            if len(imports) > 0:  # Step 3: Write new content
                imports = removeDuplicates(imports)
                imports = "\n".join(imports)
                file.write(imports + "\n")

            if code is not None:
                # TODO: we might need to rearrange chunks
                code = [chunk for chunk in code if chunk is not None]
                file.write("\n\n")
                code = removeDuplicates(code)
                code = "\n\n".join(code)
                file.write(code)  # Append old content to new content
    else:
        with open(dest,"w") as file:
            if(len(imports)>0):
                file.write("\n".join(imports))
                file.write("\n\n")
            if code is not None:
                file.write(code)

    # # TODO : sort code, problem with sorting in views.py
    if "urls.py" not in dest and "views.py" not in dest:
        sortFile(dest)


def extract_settings_references(code):
    pattern = r'settings\.\w+'
    confs = [i.replace("settings.","").strip() for i in re.findall(pattern, code)]

    return confs

def checkIfFeatureAlreadyInFile(file_path,feature):
    if os.path.exists(file_path) == False:
        return False
    
    with open(file_path) as file:
        chunks = parser.parse_code(file.read())
        names = []
        for chunk in chunks:
            chunk_name = get_assigned_variables(chunk,True)
            if len(chunk_name) > 0:
                names.append(chunk_name)

        # print("checking ",file_path," ",names, " feture ", feature)
        return feature in names
    
def get_temp_file_path(file_path, project_root,project_name,app_name):
    # print("working with ", file_path)
    file_path_list = file_path.split("/")
    file = file_path_list.pop().strip()

    main_path = "/".join(file_path_list)
    
    parent_dir = file_path_list.pop()

    main_path = main_path.split("/")[-1].strip()

    # print(app_name, " here ", parent_dir)
    # print(main_path, "  pth ",project_name)

    app_name = app_name.strip()
    parent_dir = parent_dir.strip()

    if project_name.strip() == main_path.strip():
        # TODO : convert this to write to root
        templtae_file_path = file
        

    elif app_name != parent_dir:
        # check if .sb_utils folder exist
        # if not create folder
        templtae_file_path = ".sb_utils/"+file

    else:
        templtae_file_path = "sb_app/"+file

    return templtae_file_path

# get template conf from OneStep       
def OneStep(
        file_path,
        feature,
        project_root,
        folders_in_project_root,
        writeToFile=True,
        app_name="/",
        use_email=False,
        settings_conf=[],
        project_name=None,
        output_dir = None,
        processed = []
        ):
    
    # feature_name = feature
    # print("running for feature ", feature, ' processed ', processed)
    packages = set()
    use_email = use_email
    # settings_conf = [] #change to ordered set

    # print("output dir from one fiel is ", output_dir)

    if os.path.exists(file_path):
        with open(file_path,"r") as file:
            data = file.read()
            file_dependencies = PythonBlockParser().parse_code(data)

            for chunk in file_dependencies:
                if chunk.startswith("import ") or chunk.startswith("from "):
                    if "django.core.mail" in chunk:
                        use_email = True

            # get feature
            feature_code = getCodeBlockFromFile(feature,file_dependencies)

            # get feature dependencies
            feature_dependencies = getBlockDependencies(feature_code,file_dependencies)

            # print("project path ", folders_in_project_root)

            # get feature imports
            deb_imports = [] #[f"from {e['packagePath']} import {e['imports']}" for e in feature_dependencies if e['packagePath'] != "."]
            
            for entry in feature_dependencies:
                package_path = entry['packagePath'].strip()
                imports = entry['imports'].strip()

                if package_path == imports:
                    deb_imports.append(f"import {imports}")
                elif package_path != ".":

                    # replace import here
                    # if import is coming from a custom package, replace path to .sb_utils
                    package_path_words = package_path.split(".")
                    file_name = package_path_words[-1]

                    if package_path_words[0] in folders_in_project_root:  #get all folders name in django project root folder
                        new_package_path = f".sb_utils.{file_name}" #change path here maybe use an absolute path
                        deb_imports.append(f"from {new_package_path} import {imports}")
                    else:
                        deb_imports.append(f"from {package_path} import {imports}")

            # This remove folder name from import and replace with .
            # we will unpack all of this into a sinlge django app
            # so we want to use . to refrence the app that we will unpack to as the parent dir
            
            for index,importLine in enumerate(deb_imports):
                line = importLine.split("import")[0]
                line = line.replace("from","").strip()

                if len(line) > 0 and line.startswith(".") == False:
                    folder_name = line.split(".")[0]
                    if folder_name in folders_in_project_root:
                        importLine = importLine.replace(line,line[len(folder_name):])

                        deb_imports[index] = importLine

            if writeToFile:
                template_file_path = get_temp_file_path(file_path,project_root,project_name,app_name)

                # write code to output
                # make write optional
                # add to processed here
                # print("writting  ",feature,"\n\n")
                writeCodeToFile(template_file_path,feature_code, deb_imports,output_dir=output_dir)
            
            if len(feature_dependencies) > 0:
                # loop through dependencies
                for feature_dep in feature_dependencies:
                    path,dep = [feature_dep['packagePath'], feature_dep['imports']]

                    # print("dependencies ",path," ",dep)
                    
                    if path != ".":
                        if path.startswith("."):
                            new_path = file_path.split("/")
                            new_path.pop()
                            new_path.append(f'{path[1:]}.py')
                            new_path = "/".join(new_path)
                            path = new_path
                        else:
                            package = path.split(".")[0]

                            if package != "django":
                                if package not in folders_in_project_root:
                                    packages.add(package)
                            else:
                                full_import = path + f".{dep}"

                                if full_import.startswith("django.conf.settings"):
                                    # print("found settings configuration ")
                                    # feature_code
                                    new_settings_conf = extract_settings_references(feature_code)
                                    settings_conf.extend(new_settings_conf)

                                # check for django settings import here
                                
                            path = path.split(".")
                            path = project_root + "/"+"/".join(path)+".py"

                        # if dep not in processed:
                        check_file_path = get_temp_file_path(path,project_root,project_name,app_name)
                        dest = getWritePath(check_file_path,output_dir)
                        if checkIfFeatureAlreadyInFile(dest,dep) == False:
                            feature_packages,use_email,settings_conf = OneStep(path,dep,project_root,folders_in_project_root,True,app_name,use_email,settings_conf, project_name,output_dir,processed)
                            packages = packages.union(feature_packages)
                    else:
                        # if dep not in processed:
                        check_file_path = get_temp_file_path(file_path,project_root,project_name,app_name)
                        dest = getWritePath(check_file_path,output_dir)

                        if checkIfFeatureAlreadyInFile(dest,dep) == False:
                            feature_packages, use_email, settings_conf = OneStep(file_path,dep,project_root,folders_in_project_root,True,app_name,use_email,settings_conf,project_name,output_dir,processed)
                            packages = packages.union(feature_packages)

    # print("processed : ", processed, "\n\n\n ##############################################")
    return [packages, use_email, settings_conf]

def getOldFile(filePath):
    with open(filePath) as file:
        return file.read()


# TODO: what is this function doing : its updating the actual django project file
# def writeToFile(filePath,content,fileName):

#     dest = filePath

#     if fileName not in filePath:
#         dest = f"{filePath}/{fileName}"

#     if os.path.exists(dest):  

#         data = getOldFile(dest)

#         if "urls.py" == fileName:
#             file_content = merge_urls(data.split("\n"),content.split("\n"))

#         else:
#             # TODO : handle conflicts here
#             write_data = f"{content}\n\n{data}" # node seperate with two \n e get Y

#             code = []
#             imports = []
#             file_content = ""
#             chunks = PythonBlockParser().parse_code(write_data)
            
#             for chunk in chunks:
#                 if chunk.startswith("import ") or chunk.startswith("from "):
#                     # individualImports = getIndividualImports(chunk)
#                     imports.append(chunk)
#                 else:
#                     code.append(chunk.strip())

#             if imports:  # Step 3: Write new content
#                 imports = removeDuplicates(imports)
#                 imports = "\n".join(imports)
#                 file_content += f"{imports} \n\n\n"

#             code = removeDuplicates(code)
#             code = "\n\n".join(code)

#             file_content += code  # Append old content to new content

#         with open(dest,"w") as file:
#             file.write(file_content)

#     else:
#         with open(dest,"w") as file:
#             file.write(content)

def getFileImportsAndCode(chunks):
    imports = []
    code = []
    
    for chunk in chunks:
        if chunk.startswith("import ") or chunk.startswith("from "):
            # individualImports = getIndividualImports(chunk)
            imports.append(chunk)
        else:
            code.append(chunk.strip())

    return [imports,code]

def writeToFile(filePath,content,fileName,conflicts=[]):

    dest = filePath

    if fileName not in filePath:
        dest = f"{filePath}/{fileName}"

    if os.path.exists(dest):  

        currentFileContent = getOldFile(dest)

        if "urls.py" == fileName:
            file_content = merge_urls(currentFileContent.split("\n"),content.split("\n"))

        else:
            # TODO : handle conflicts here
            file_content = ""

            currentFileChunks = parser.parse_code(currentFileContent)
            codeUpdateChunk = parser.parse_code(content) #new update from sb deploy
            
            imports, currentFileCode = getFileImportsAndCode(currentFileChunks)
            newImports, newCode = getFileImportsAndCode(codeUpdateChunk)
            
            # get chunks name from currentFileCode
            currentFileChunksNames = []
            for chunk in currentFileCode:
                name = get_assigned_variables(chunk,True)
                if isinstance(name,str):
                    currentFileChunksNames.append(name)
                else:
                    # add a default value just to retain chunk order
                    currentFileChunksNames.append(0)

            # process new code and decide on insert position
            # default insert position is 0
            for chunk in newCode:
                name = get_assigned_variables(chunk,True)
                if isinstance(name,str):
                    if name in currentFileChunksNames:
                        """ Conflict Identified """
                        conflictChunkPosition = currentFileChunksNames.index(name) #assuming name appears only once
                        oldCode = currentFileCode[conflictChunkPosition]

                        if oldCode != chunk:
                            # pop code
                            currentFileCode.pop(conflictChunkPosition)

                            # add conflict markers
                            conflictCode = f"<<<<<<< SpeedBuild update \n{oldCode}\n=======\n{chunk}\n>>>>>>>"
                            currentFileCode.insert(conflictChunkPosition, conflictCode)

                            conflicts.append(f"Merge conflict in {dest} : Failed to merge {name}")

                        continue

                currentFileCode.insert(0,chunk) # fresh non merge conflict code

            imports.extend(newImports) # merge current and new imports

            if imports:  # Step 3: Write new content
                imports = removeDuplicates(imports)
                imports = "\n".join(imports)
                file_content += f"{imports} \n\n\n"

            code = removeDuplicates(currentFileCode)
            code = "\n\n".join(code)

            file_content += code  # Append old content to new content

        with open(dest,"w") as file:
            file.write(file_content)

    else:
        with open(dest,"w") as file:
            file.write(content)

    return conflicts


def sortFile(file_name):
    with open(file_name,"r") as file:
        data = file.read()
        codeChunks = []
        imports = []

        chunks = PythonBlockParser().parse_code(data)

        for chunk in chunks:
            if chunk.startswith("import ") or chunk.startswith("from "):
                imports.append(chunk)
            else:
                codeChunks.append(chunk)

        arranged_chunks = []
        processed = []
    
    file_imports = "\n".join(imports)
    processedChunks = arrangeChunks(codeChunks,arranged_chunks,processed)
    file_code = "\n\n".join(processedChunks)

    with open(file_name,"w") as sortedFile:
        newData = f"{file_imports}\n\n{file_code}"
        sortedFile.write(newData)


import re
import os
import yaml
import shutil

from pathlib import Path

from ...utils.template_update import checkTemplateForVersionUpdate

from ...src.Django.utils.django_utils import ManageDjangoSettings
from ...src.Django.utils.extract import clear_folder, extract_feature_code_and_dependencies, getTemplateFileNames, getURLForFeature
from ...src.Django.utils.feature_dependencies import arrangeChunks
from ...src.Django.utils.venv_utils import get_activated_venv
from ...utils.compare_templates import compare_zip_code_folders
from ...utils.utils import findFilePath, get_template_output_path

# remove comments from code before splitting blocks
def create_temp_from_feature(project_path,project_name,feature_name,feature_file_path,template_django_apps):

    venv = get_activated_venv()
    

    # Get installed packages and their versions from venv
    # packages = {}
    # if venv:
    #     dist_packages = pkg_resources.working_set
    #     for package in dist_packages:
    #         packages[package.key] = package.version

    # print("packages are ", packages)
    
    # get settings conf from django.conf.settings TODO
    relative_path = feature_file_path.replace(project_path,"").split("/")
    relative_path.pop()
    app_name = relative_path.pop()


    findFilePath(project_path, "urls.py")

    # use conditional statement to check if file is found
    project_name = findFilePath(project_path, "asgi.py")

    if len(project_name) == 0:
        raise(ValueError("Cannot find django project settings file"))
    
    template_dep = set()
    installed_apps = set()
    project_name = project_name[0].split("/")[0]
    settings_path = f"{project_name}/settings.py"
    settings_path = os.path.join(project_path, settings_path)

    project_dir = [folder for folder in os.listdir(project_path) if os.path.isdir(f"{project_path}/{folder}")]
    
    # here
    # template_django_apps = getDjangoAppsPackage(settings_path,venv)

    settings = ManageDjangoSettings(settings_path,venv)

    processed = set()


    # start; make recursive

    output_folder_name = "sb_output_"+feature_name

    template_settings_import,template_confs,template_dep = extract_feature_code_and_dependencies(
        feature_file_path,
        feature_name,
        project_path,
        project_dir,
        app_name,
        template_django_apps,
        settings,
        project_name,
        installed_apps,
        template_dep,
        processed,
        [],
        None,
        output_folder_name
    )

    # end recursion
    output_dir = "./output/"+output_folder_name

    # Create template yaml file
    template_confs = arrangeChunks(template_confs,[],[])
    
    data = {
        "feature" : feature_name,
        "feature_file_path" : feature_file_path.replace(project_path,""),
        "source_project":project_name,
        "dependencies" : list(template_dep),
        "settings" : {
            "imports" : list(sorted(set(template_settings_import))),
            "installed_apps" : list(installed_apps),
            "middlewares" : settings.getTemplateMiddleWare(list(installed_apps)),
            "configurations" : template_confs#list(set(template_confs))
        }
    }

    # print("template configurations ",template_confs)

    with open(f'{output_dir}/sb_{feature_name}.yaml', 'w') as file:
        yaml.dump(data, file, default_flow_style=False,sort_keys=False)


    if os.path.exists(f"{output_dir}/sb_app/views.py"):
        # if views.py in models, generate url for feature
        getURLForFeature(feature_name,feature_file_path,app_name,project_path,output_folder_name)

    # create template documentations here
    tem_files = getTemplateFileNames(output_dir)
    tem_files = sorted(tem_files, key=lambda x: not x.startswith('.sb_utils')) #process dependencies first
    # Exclude the yaml file

    # for file in tem_files:
    #     file_name = file.split("/")
    #     main_file = file_name.pop().split(".")[0] #remove extention from file name
    #     file_name.append(main_file)
    #     file_name = "_".join(file_name)
  
    #     main_dir = "./output/documentation"
    #     if not os.path.exists(main_dir):
    #         os.makedirs(main_dir)

    #     # write to file  TODO : leave commented for now
    #     with open(f"{main_dir}/{file_name}.md","w") as new_file:
    #         # get and write doc to file
    #         pass
    #         # doc = documentationAgent(file)
    #         # new_file.write(doc)

    # Save template to user main dir in the .sb_zip folder
    user_dir = str(Path.home())
    user_dir += "/.sb_zip"

    if not os.path.exists(user_dir):
        os.makedirs(user_dir)

    # print("Saving to ", user_dir)

    # output_zip = f"{user_dir}/speed_build_{feature_name}"  # No .zip extension needed
    template_path = get_template_output_path(feature_name)

    shutil.make_archive(template_path, 'zip', output_dir)


    # delete output folder
    clear_folder(output_dir)
    
    # structure templates to include versions
    checkTemplateForVersionUpdate(template_path)

    print(f"Template `{template_path}` Created")




import json
import sys

from .auth import manageAuth
from .cli_utils import setAPIKey,updateDefaultLLMModel,get_config

supported_models = ["gpt-4o-2024-08-06","sonnet-3.7"]
supported_providers = ["openai","anthropic","google"]

async def sbAuth():
    # remove auth keyword
    sys.argv.pop(1)
    command = None

    try:
        command = sys.argv[1]
        if command == "llm" and len(sys.argv) > 2:
            subcommand = sys.argv[2]
            
            if subcommand == "provider":
                if len(sys.argv) == 5:
                    provider = sys.argv[3].lower()
                    api_key = sys.argv[4]

                    if provider in supported_providers:
                        print("setting provider for provider ",provider," with api key ",api_key)
                        setAPIKey(provider,api_key)
                    else:
                        print("Please choose form our LLM supported providers ",supported_providers)
                else:
                    print("speedbuild llm provider <provider_name> <provider_api_key>")

            elif subcommand == "model":
                if len(sys.argv) == 4:
                    model = sys.argv[3].lower()
                    if model in supported_models:
                        print("setting default model to ", model)
                        updateDefaultLLMModel(model)
                    else:
                        print("choose from our supported models : ",supported_models)
                else:
                    print("speedbuild llm <model_name>")


        elif command == "config":
            get_config()

        elif command == "login":
            print("logining in user")
            await manageAuth()
        elif command == "register":
            await manageAuth("register")
        else:
            raise IndexError
        
    except IndexError:
        if command is not None and command == "llm":
            print("Please Enter a valid action for llm")
            print("  - speedbuild auth llm model        :\tManage LLM model")
            print("  - speedbuild auth llm provider     :\tManage LLM Provider Access")
        else:
            print("Please Enter a valid action for auth")
            print("  - speedbuild auth config      :\tSee SpeedBuild Config and API Keys")
            print("  - speedbuild auth login       :\tAuthenticate with server")
            print("  - speedbuild auth register    :\tCreate new SpeedBuild account")
            print("  - speedbuild auth llm         :\tManage LLM Configurations")


new_system_prompt = """
You are a **smart AI agent** proficient in **Django**. 
You are part of a service called **SpeedBuild**, 
which helps developers extract features from their code, 
customize those features using plain English commands, and reuse them in new projects.

A feature and its dependencies may span multiple files. Your job is to:

1. **Process dependencies first** before modifying the main feature.
2. **Retrieve file contents when needed** by requesting the user to provide them.
3. **Modify the necessary code while maintaining structure and clarity.**
4. **Ensure all code modifications include meaningful comments** to aid understanding.
5. **Return responses in a structured JSON format** to allow for easy parsing by a Python script.
6. **Ensure that you go through every file listed by the user** to determine whether modifications are needed.
7. **If a file does not need modifications, respond with "copy" and the file name.**
8. **If a file needs to be deleted, respond with "delete" and the file name.**
9. **Terminate the session only when all files have been reviewed and no further actions are required.**
10. **NOTE: When creating a new file, always prepend the folder name to the file name. The only exception is when creating a file in the root folder.**

---

### How You Should Respond

Always return a structured JSON response with the following top-level fields:

* **`status`**:

  * `"request_file"` → When you need the content of a file.
  * `"success"` → When you have successfully processed and modified a file.
  * `"delete"` → When a file should be deleted.
  * `"copy"` → When a file remains unchanged.
  * `"done"` → When all necessary actions are complete.
* **`action`**: A short description of the action you are taking.
* **`file_name`**: The name of the file being referenced or changed.
* **`content`**: Varies depending on the `status` — see formats below.
* **`comments`**: A brief explanation of the changes made or why the file was not modified.

---

### Code Block Update Format (when `status` is `"success"`)

When modifying a file, you must return a list of code block instructions inside the `content` field. Each item must contain:

```json
{
  "action": "add" | "replace" | "remove",
  "new_code": "<code to add or replace with>",
  "old_code": "<only for 'replace' or 'remove'>",
  "comment": "A short explanation of the change"
}
```

* If `action` is `"add"`: Only `new_code` is required.
* If `action` is `"replace"`: Both `new_code` and `old_code` are required.
* If `action` is `"remove"`: Only `old_code` is required.

---

### Response Examples

#### Requesting a File

```json
{
  "status": "request_file",
  "action": "Requesting file content",
  "file_name": "sb_app/models.py"
}
```

---

#### Modifying a File

```json
{
  "status": "success",
  "action": "Added 'Chef' model and updated 'Recipe' model to include 'prepared_by'",
  "file_name": "sb_app/models.py",
  "content": [
    {
      "action": "add",
      "new_code": "class Chef(models.Model):\n    name = models.CharField(max_length=100)\n    bio = models.TextField(blank=True)\n\n    def __str__(self):\n        return self.name",
      "comment": "Added a new 'Chef' model"
    },
    {
      "action": "replace",
      "old_code": "class Recipe(models.Model):\n    name = models.CharField(max_length=100)\n    directions = models.TextField()",
      "new_code": "class Recipe(models.Model):\n    name = models.CharField(max_length=100)\n    directions = models.TextField()\n    prepared_by = models.ManyToManyField(Chef, blank=True)",
      "comment": "Linked Recipe to Chef using ManyToManyField"
    }
  ],
  "comments": "Added a new Chef model and linked it to the Recipe model for tracking who prepared each recipe."
}
```

---

#### Registering Models in `admin.py`

```json
{
  "status": "success",
  "action": "Registered 'Recipe' and 'Chef' models in Django admin",
  "file_name": "home/admin.py",
  "content": [
    {
      "action": "replace",
      "old_code": "from .models import Recipe",
      "new_code": "from .models import Recipe, Chef",
      "comment": "Updated import statement to include Chef model"
    },
    {
      "action": "add",
      "new_code": "admin.site.register(Chef)",
      "comment": "Registered Chef model with Django admin"
    }
  ],
  "comments": "Updated import and registered Chef model for admin access."
}
```

---

#### Deleting a File

```json
{
  "status": "delete",
  "action": "Remove unused test file",
  "file_name": "tests/temp_feature.py",
  "comments": "This file was used for temporary debugging and is no longer needed."
}
```

---

#### No Change Needed

```json
{
  "status": "copy",
  "action": "No changes required",
  "file_name": "utils/helpers.py",
  "comments": "The helper functions are already compatible with the new feature."
}
```

---

#### Ending Session

```json
{
  "status": "done",
  "action": "All tasks completed",
  "file_name": "",
  "content": "",
  "comments": "Reviewed and updated models.py, admin.py, and views.py. Registered new model and updated checkout logic."
}
```

"""

system_prompt2 = """
### **Updated Prompt**  

You are a **smart AI agent** that is proficient in **Django**. You are part of a service called **SpeedBuild**, which helps developers extract features from their code, customize those features using plain English commands, and reuse them in new projects.  

A feature and its dependencies may span multiple files. Your job is to:  
1. **Process dependencies first** before modifying the main feature.  
2. **Retrieve file contents when needed** by requesting the user to provide them.  
3. **Modify the necessary code while maintaining structure and clarity.**  
4. **Ensure all code modifications include meaningful comments** to aid understanding.  
5. **Return responses in a structured JSON format** to allow for easy parsing by a Python script.  
6. **Ensure that you go through every file listed by the user** to determine whether modifications are needed.  
7. **If a file does not need modifications, respond with "copy" and the file name.**  
8. **If a file needs to be deleted, respond with "delete" and the file name.**  
9. **Terminate the session only when all files have been reviewed and no further actions are required.**  
10. **NOTE : when creating always prepend the folder to create the file in to the file name. **

---

### **How You Should Respond**  

You **must always** return a structured JSON response with the following fields:  

- **`status`**:  
  - `"request_file"` → When you need the content of a file.  
  - `"success"` → When you have successfully processed and modified a file.  
  - `"delete"` → When a file is no longer needed and should be removed.  
  - `"done"` → When all necessary modifications and file requests are complete, and no further actions are required.  
- **`action`**: A short description of what you are doing.  
- **`file_name`**: The name of the file being processed or requested.  
- **`content`**: The modified or newly generated code (if applicable).  
- **`comments`**: A brief explanation of the changes made for clarity.  

---

### **Examples**  

#### **1. Requesting a File**  
When you need a file’s content, return:  

```json
{
  "status": "request_file",
  "action": "Requesting file content",
  "file_name": "sb_app/models.py"
}
```

#### **2. Updating a File**  
Once you have the file, modify it and return:  

```json
{
  "status": "success",
  "action": "Added a 'Chef' model and linked it to 'Recipe' via a ManyToManyField",
  "file_name": "sb_app/models.py",
  "content": "from django.db import models\n\n# Defining the Chef model\nclass Chef(models.Model):\n    name = models.CharField(max_length=100)  # Stores the chef's name\n    bio = models.TextField(blank=True)  # Optional field for additional information\n\n    def __str__(self):\n        return self.name\n\n# Updating the Recipe model\nclass Recipe(models.Model):\n    name = models.CharField(max_length=100)\n    image = models.FileField(null=True, default=None)\n    description = models.TextField()\n    nutritional_description = models.TextField()\n    ingredients = models.TextField()\n    directions = models.TextField()\n    clap = models.PositiveIntegerField(default=0)\n    tags = models.ManyToManyField('Tag')\n    meta = models.TextField(blank=True)\n    options = models.ManyToManyField('RecipeOption', blank=True)\n    prepared_by = models.ManyToManyField(Chef, blank=True)  # Linking Recipe to Chef\n\n    def __str__(self):\n        return self.name",
  "comments": "Added a 'Chef' model to store chef details. Updated 'Recipe' to include a ManyToManyField linking it to 'Chef'."
}
```

#### **3. Registering Models in `admin.py`**  

```json
{
  "status": "success",
  "action": "Registered 'Recipe' and 'Chef' models in the Django admin panel",
  "file_name": "home/admin.py",
  "content": "from django.contrib import admin\nfrom .models import Recipe, Chef\n\n# Registering models in Django admin\nadmin.site.register(Recipe)\nadmin.site.register(Chef)",
  "comments": "Registered 'Recipe' and 'Chef' in Django's admin panel to allow management through the admin interface."
}
```

### **4. Ending session**
```json
{
  "status": "done",
  "action": "All tasks completed",
  "file_name": "",
  "content": "",
  "comments": <a short descriotion of all the work done>
}
```

---

### **Key Instructions for the AI Agent**  
1. **Always process dependencies before the main feature.**  
2. **Request file contents if they are not provided.**  
3. **Modify only what is necessary** and **preserve existing code.**  
4. **Ensure all code includes meaningful comments.**  
5. **Respond with a structured JSON output** so a Python script can process it.  
6. **always specify full path to file during creation the only exception is when you want to create a file in the root folder**

"""


from ..auth.cli_utils import get_sb_config


openai = ["gpt-4o-2024-08-06"]
# anthropic = []

def getLLMConnectInfo():
    config = get_sb_config()
    if "default_model" not in config.keys() and "llm_keys" not in config.keys:
        return None
    
    model = config['default_model']
    llm_keys = config['llm_keys']

    if model in openai:
        if "openai" in llm_keys.keys():
            return ["openai",llm_keys['openai'],model]
        
        print("Sent OpenAI API key to proceed")
        
    else:
        return 0


import os
from pathlib import Path

from .src.Django.deploy_features import convertFromTemplateToFeature
from .utils.utils import getAbsolutePath
# import sys
# import asyncio
# from pathlib import Path

# from .src.Django.deploy_features import convertFromTemplateToFeature
from .src.Django.extract_features import create_temp_from_feature

# from .auth.auth import manageAuth
# from .utils.pull_template import pullInitialTemplates

# from .auth.sb_user import sbAuth
# from .src.undo import undoSBDeploy
# from .auth.serverCall import pingServer
# from .utils.package_utils import getPackageNameMapping
# from .utils.utils import extract_views_from_urls, findFilePath, getAbsolutePath, pullPythonPackageJSON

# def createTemplate(file_path, feature,project_name,packageToNameMapping,project_root):

#     if file_path.endswith("/"):
#         file_path = file_path[:len(file_path)-1]

#     file_path = file_path.replace(project_root,"")

#     print("\n##### Extracting Feature : ",feature," #####")

#     if not os.path.exists(file_path):
#         print("invalid feature")
#         return

#     create_temp_from_feature(
#         project_root,
#         project_name,
#         feature,
#         file_path,
#         packageToNameMapping
#     )

# def implementFeature(template_path,project_path=None,appName=None,prompt=None, addToInstalledApps=None,venv=None):
#     project_root = getAbsolutePath(".")

#     if project_path:
#         project_root = project_path

#     if ":" in template_path:
#         # User specified feature version
#         template_path_words = template_path.split(":")

#         if len(template_path_words) == 2:
#             template_name,version = template_path_words
#             template_path = f"{template_name}/{version}.zip"
#         else:
#            print("invalid template version")
#            return 
#     else:
#         # version was not specified, default to lts
#         template_name = template_path
#         template_path += "/lts.zip"

#     home = str(Path.home())
#     template_path = f"{home}/.sb_zip/{template_path}"

#     print("finished processing ",template_path, " template name is ", template_name)

#     if not os.path.exists(template_path):
#         print("invalid template")
#         return
    
#     # template_path = getAbsolutePath(args.template_path)
#     # template_name = template_path.split("/")[-1].replace(".zip","").strip()
#     # just commented tihs out; Investigate this
#     # extract_path = f"{home}/.sb/sb_extracted/{template_name}"
    
#     print("deploying feature")
#     convertFromTemplateToFeature(project_root,template_path,template_name,appName,prompt,addToInstalledApps,venv)

# def list_features():
#     user_home = str(Path.home())
#     templates_dir = f"{user_home}/.sb_zip"
#     if not os.path.exists(templates_dir):
#         os.makedirs(templates_dir)
#     templates = os.listdir(templates_dir)
#     count = 1
#     print("## ALL EXTRACTED FEATURES ##\n")
#     for template in templates:
#         if template.startswith("speed_build"):
#             print(f"    {count}) {template}")
#             count += 1
#     print("\n")


# def handleExtract(subfolder=None,feature=None,project_path=""):
#     project_root = project_path
#     # get project name from root project path
#     project_name = [i.strip() for i in project_path.split("/") if len(i.strip()) > 0][-1]

#     if subfolder:
#         project_path += subfolder

#     if not project_path.endswith("/"):
#         project_path += "/"

#     if feature is not None:
#         # print("here with",feature)
#         if project_path.endswith("/"):
#             project_path = project_path[:len(project_path)-1]

#         # get package to name mapping
#         packageToNameMapping = getPackageNameMapping(project_root)
#         print("\n#####  Extracting ",feature, "from ",project_path,"  #####\n")
#         createTemplate(project_path, feature,project_name,packageToNameMapping,project_root)
#         return
    
#     elif subfolder and feature == None:
#         if project_path.endswith("/"):
#             project_path = project_path[:len(project_path)-1]

#         if project_path.endswith(".py"):
#             paths = project_path.split("/")
#             paths.pop()
#             project_path = "/".join(paths)

#         if not project_path.endswith("/"): 
#             project_path += "/"

#         # print(project_path, " new edited path")

#     urls_path = findFilePath(project_path,"urls.py")


#     packageToNameMapping = getPackageNameMapping(project_root)
#     # print(packageToNameMapping, " for app")

#     for filePath in urls_path:
#         print(f"\n{"#"*15} {filePath} {"#"*15}\n")
#         abs_path = project_path + filePath
#         views = extract_views_from_urls(abs_path)

#         dir_name = filePath.split("/")[0]

#         print("directory name is ",dir_name)
#         # return

#         for view in views:
#             view = view.split(".")

#             print("view list is ",view)

#             if len(dir_name) == 0:
#                 dir_name = "."

#             file_name = view[1]
#             feature_name = view[2]

#             # TODO Fix : we are assuming urls path are like :
#             # path("home",views.home,name="home")
#             # we are assuming the user is import views like 
#             # from . import views

#             # if feature_name.endswith("/"):
#             #     feature_name = feature_name[:len(feature_name)-1]

#             # print("feature is ", feature_name)
#             print("here we are")
#             print(f"folder : {dir_name} file : {file_name}.py feature is {feature_name} ",project_path)
#             createTemplate(project_path+f"{dir_name}/views.py", feature_name,project_name,packageToNameMapping,project_root)


# def main(action,secPath,feature,project_path):
#     if action == "list":
#         list_features()
    
#     elif action == "extract":
#         handleExtract(secPath,feature,project_path)
    
#     elif action == "deploy":
#         if secPath is not None:
#             # print("deploying ", secPath)
#             implementFeature(secPath)
#         else:
#             print("Specify a template to deploy\n")
#             print("  You can see all extracted templates by running")
#             print("  sb.py list ")
    # elif action == "undo":
    #     print("performing undo")
    #     undoSBDeploy()
    # elif action == "auth":
    #     asyncio.run(sbAuth())
    # elif action == "ping":
    #     pingServer()
    # elif action == "setup":
    #     print("Seting up speedbuild")
    #     pullInitialTemplates()
    #     pullPythonPackageJSON()
    #     asyncio.run(manageAuth("register"))
#     else:
#         print("Please Enter a valid action")
#         print("  - speedbuild setup   :\tInitial Setup")
#         print("  - speedbuild extract :\tto extract templates")
#         print("  - speedbuild list    :\tto see extracted templates")
#         print("  - speedbuild deploy  :\tto deploy template to a different project")
#         print("  - speedbuild auth    :\tto manage authentication and LLM configuration")


# def start():
#     project_path = getAbsolutePath(".")
#     if not project_path.endswith("/"):
#         project_path += "/"

#     args = sys.argv[1:] + [None] * 3  # Pad with None for missing values
#     action, secondaryPath, feature = args[:3]

#     main(action, secondaryPath, feature, project_path)

# if __name__ == "__main__":
#     start()
    
def createTemplate(file_path, feature,project_name,packageToNameMapping,project_root):

    if file_path.endswith("/"):
        file_path = file_path[:len(file_path)-1]

    file_path = file_path.replace(project_root,"")

    print("\n##### Extracting Feature : ",feature," #####")

    if not os.path.exists(file_path):
        print("invalid feature")
        return

    create_temp_from_feature(
        project_root,
        project_name,
        feature,
        file_path,
        packageToNameMapping
    )


def implementFeature(template_path,project_path=None,appName=None,prompt=None, addToInstalledApps=None,venv=None):
    project_root = getAbsolutePath(".")

    if project_path:
        project_root = project_path

    if ":" in template_path:
        # User specified feature version
        template_path_words = template_path.split(":")

        if len(template_path_words) == 2:
            template_name,version = template_path_words
            template_path = f"{template_name}/{version}.zip"
        else:
           print("invalid template version")
           return 
    else:
        # version was not specified, default to lts
        template_name = template_path
        template_path += "/lts.zip"

    home = str(Path.home())
    template_path = f"{home}/.sb_zip/{template_path}"

    print("finished processing ",template_path, " template name is ", template_name)

    if not os.path.exists(template_path):
        print("invalid template")
        return
    
    # template_path = getAbsolutePath(args.template_path)
    # template_name = template_path.split("/")[-1].replace(".zip","").strip()
    # just commented tihs out; Investigate this
    # extract_path = f"{home}/.sb/sb_extracted/{template_name}"
    
    print("deploying feature")
    convertFromTemplateToFeature(project_root,template_path,template_name,appName,addToInstalledApps,venv)


import os
import uuid
import json
import shutil
from pathlib import Path

def getAllProjectId():
    home = str(Path.home())
    path_to_sb_state = home + "/.sb/state"
    if not os.path.exists(path_to_sb_state):
        os.makedirs(path_to_sb_state,exist_ok=True)

    # get all projects
    allProjects = os.listdir(path_to_sb_state)
    return allProjects

def getProjectPathFromId(id):
    home = str(Path.home())
    project_path = home + "/.sb/state/"+id

    return project_path

def getInitialStateContent():
    return {
        "action":{},
        "undo":False
    }

def readStateJSON(project_id):
    project_path = getProjectPathFromId(project_id)
    state_path = project_path + "/state.json"

    if not os.path.exists(state_path):
        os.makedirs(project_path,exist_ok=True)
        with open(state_path,"w") as file:
            json.dump(getInitialStateContent(), file)

    with open(state_path,"r") as file:
        data = json.loads(file.read())
        return data

def updateState(project_id,action,filename):
    state = readStateJSON(project_id)

    project_path = getProjectPathFromId(project_id)
    state_path = project_path + "/state.json"

    actions = state['action']
    actions[filename] = action

    undo = state['undo']

    with open(state_path, "w") as file:
        json.dump({
            "action": actions,
            "undo": undo
        }, file, indent=4)

def generateNewProjectId():
    all_ids = set(getAllProjectId()) 
    for _ in range(1000):
        new_id = uuid.uuid4().hex[:16]
        if new_id not in all_ids:
            return new_id
    # Fallback – extremely rare
    return f"{new_id}:c{list(all_ids).count(new_id)}"

def clearProjectState(project_id):
    project_path = getProjectPathFromId(project_id)

    if os.path.exists(project_path):
        shutil.rmtree(project_path)

    # create fresh state.json file
    state_path = project_path + "/state.json"
    os.makedirs(project_path, exist_ok=True)
    with open(state_path,"w") as file:
        json.dump(getInitialStateContent(), file)

def getOrCreateProjectSBId(project_path):
    if not project_path.endswith("/"):
        project_path += "/"
    speedbuild_file_path = project_path + ".speedbuild"

    if os.path.exists(speedbuild_file_path):
        with open(speedbuild_file_path) as file:
            sb_id = file.read()
    else:
        # generate new id
        sb_id = generateNewProjectId()

        # create new .speedbuild file in project root
        with open(speedbuild_file_path,"w") as file:
            file.write(sb_id)

    return sb_id

def copyFileToState(project_id, source,project_root="", destination="prev"):
    sb_project_path = getProjectPathFromId(project_id)
    dest_dir = os.path.join(sb_project_path, destination)

    os.makedirs(dest_dir, exist_ok=True)

    relative_source_path = source[len(project_root):].lstrip("/")
    dest_file = os.path.join(dest_dir, relative_source_path)
    dest_file_folder = os.path.dirname(dest_file)

    if not os.path.exists(dest_file_folder):
        os.makedirs(dest_file_folder,exist_ok=True)

    if os.path.exists(source):
        shutil.copy2(source, dest_file)
    return dest_file



import os
from pathlib import Path
from openai import OpenAI
from typing import Optional, Union, List
from ..parsers.python.parser import PythonBlockParser
from pydantic import BaseModel
from ..src.Django.utils.feature_dependencies import removeDuplicates
from .customize_system_prompt import system_prompt
from ..utils.llm_utils import getLLMConnectInfo
from ..utils.update_template_code import addOrUpdateCode



memory = [
    {"role": "system", "content": system_prompt},
]

parser = PythonBlockParser()

class CodeChange(BaseModel):
    action: str
    new_code: str
    comment: Optional[str] = None
    old_code: Optional[str] = None

class SpeedBuildFormat(BaseModel):
    status: str
    action: str
    file_name: str
    # content: Optional[List[CodeChange]] = None
    content: Optional[Union[str, List[CodeChange]]] = None
    comment: Optional[str] = None


def getLLMClient():
    llm_config = getLLMConnectInfo()
    if llm_config == None:
        raise ValueError("Please add an LLM API Key to proceed with customization")
    elif llm_config == 0:
        raise ValueError("This Version of Speedbuild only support OpenAI")
    
    api_key = llm_config[1]
    model = llm_config[2]

    client = OpenAI(api_key=api_key)

    return [model,client]

def makeLLMCall(user_input):
    model,client = getLLMClient()
    memory.append({"role": "user","content": user_input})
    completion = client.beta.chat.completions.parse(
        model=model,
        messages=memory,
        response_format=SpeedBuildFormat,
    )

    # Extract the assistant's response
    assistant_response = completion.choices[0].message.parsed #completion.choices[0].message.content

    # Add the assistant's response to memory
    memory.append({"role": "assistant", "content": f"{assistant_response.file_name} : {assistant_response.action} \n {assistant_response.content}"})

    return assistant_response

def getFileContent(filePath,fileName,template_name)->str:
    dest = f"{filePath}/{fileName}"
    print("file ",dest)
    
    if os.path.exists(dest):
        with open(dest, "r") as file:
            data = file.read()
            return data 
    else:
        return "File Does Not Exist"

# Todo edit here
def writeToFile(filePath,content,fileName):
    dest = f"{filePath}/{fileName}"
    print("Saving to file ",dest)
    if os.path.exists(dest):
        # append to file
        with open(dest, "r+") as file:
            data = file.read()  # Step 1: Read existing content
        
            file.seek(0)        # Move to the beginning of the file
            file.truncate(0)    # Step 2: Clear the file

            write_data = f"{content}\n{data}"
            chunks = parser.parse_code(write_data) #split_code_into_sections(write_data)
            code = []
            imports = []
            
            for chunk in chunks:
                if chunk.startswith("import ") or chunk.startswith("from "):
                    imports.append(chunk)
                    # individualImports = getIndividualImports(chunk)
                    # imports.extend(individualImports)
                else:
                    code.append(chunk)

            if imports:  # Step 3: Write new content
                imports = removeDuplicates(imports)
                imports = "\n".join(imports)
                file.write(imports + "\n")

            file.write("\n\n")
            code = removeDuplicates(code)
            code = "\n".join(code)
            file.write(code)  # Append old content to new content
    else:
        with open(dest,"w") as file:
            file.write(content)

def agent(files,prompt,project_root,template_name):
    file_processed = set()
    read_path = project_root.split("/")
    read_path.pop()
    read_path = "/".join(read_path)
    print("read path is ", read_path)

    home = str(Path.home())
    # /home/attah/.sb/sb_extracted/speed_build_UserOnboarding/sb_app
    root_path = f"{home}/.sb/sb_extracted/speed_build_{template_name}"

    while len(file_processed) < len(files):
        old_code = prompt
        prompt = "task_description: " + prompt + "\n" + f"files : {files}"
        response = makeLLMCall(prompt)
        
        status = response.status
        file_name = response.file_name
        action = response.action

        print(f"{file_name} : {action}")
        print(f"{response.status} {response.comment} == \n\n{response.content}\n\n")
        
        if status == "request_file":
            file_content = getFileContent(root_path,file_name,template_name)
            prompt = file_content
            
        elif status == "success":
            dest = f"{read_path}/.sb/{template_name}"
            code_content = response.content
            comment = response.comment

            # TODO reedit file to replace or add new code
            if code_content != None:
                if file_name in files:
                    new_chunk = parser.parse_code(old_code)
                    for code in code_content:
                        old_code = code.old_code

                        if "sb_app" in file_name:
                            code = code.new_code.replace("sb_app.",".")
                        else:
                            code = code.new_code
                            
                        new_chunk = addOrUpdateCode(new_chunk,code, old_code)

                    edited_or_new_code = "\n".join(new_chunk)

                else:
                    edited_or_new_code = []
                    
                    for block in code_content:
                        code = block.new_code
                        if "sb_app" in file_name:
                            code = code.replace("sb_app.",".")

                        edited_or_new_code.append(block.new_code)

                    edited_or_new_code = "\n".join(edited_or_new_code)
                # if "sb_app" in file_name and code_content is not None:
                #     code_content = code_content.replace("sb_app.",".")

                # file_name = file_name.split("/")[-1]
                writeToFile(root_path,edited_or_new_code,file_name) #project_root
                print(f"{file_name} : {comment}")
                # add file_name to file_processed
                file_processed.add(file_name)
                prompt = "next"
            
        elif status == "delete":
            # delete file
            pass

        elif status == "done":
            print(f"Completed : {response.comment}")
            break #stop loop
            
        continue


import os
import sys
import venv
import subprocess
from pathlib import Path


def create_virtual_environment(venv_path, python_executable=None):
    """
    Create a virtual environment programmatically.
    
    Args:
        venv_path (str): Path where the virtual environment will be created
        python_executable (str, optional): Path to specific Python executable to use
    
    Returns:
        bool: True if successful, False otherwise
    """

    venv_path = venv_path+"/venv"

    if not os.path.exists(venv_path):
        os.makedirs(venv_path)
    try:
        # Method 1: Using the venv module (Python 3.3+)
        print(f"Creating virtual environment at: {venv_path}")
        
        # Create the venv directory if it doesn't exist
        venv_dir = Path(venv_path)
        venv_dir.parent.mkdir(parents=True, exist_ok=True)
        
        # Create virtual environment
        if python_executable:
            # Use specific Python executable
            result = subprocess.run([
                python_executable, "-m", "venv", str(venv_path)
            ], capture_output=True, text=True)
            
            if result.returncode != 0:
                print(f"Error creating venv: {result.stderr}")
                return False
        else:
            # Use current Python interpreter
            venv.create(venv_path, with_pip=True)
        
        print(f"Virtual environment created successfully at: {venv_path}")
        
        # Get activation script path
        if sys.platform == "win32":
            activate_script = os.path.join(venv_path, "Scripts", "activate.bat")
            pip_path = os.path.join(venv_path, "Scripts", "pip.exe")
        else:
            activate_script = os.path.join(venv_path, "bin", "activate")
            pip_path = os.path.join(venv_path, "bin", "pip")
        
        print(f"Activation script: {activate_script}")
        print(f"Pip executable: {pip_path}")
        
        return True
        
    except Exception as e:
        print(f"Error creating virtual environment: {e}")
        return False
    

def install_packages_in_venv(venv_path, packages):
    """
    Install packages in the virtual environment.
    
    Args:
        venv_path (str): Path to the virtual environment
        packages (list): List of package names to install
    
    Returns:
        bool: True if successful, False otherwise
    """
    try:
        # Get pip path based on OS
        if sys.platform == "win32":
            pip_path = os.path.join(venv_path, "Scripts", "pip.exe")
        else:
            pip_path = os.path.join(venv_path, "bin", "pip")
        
        # Install packages
        for package in packages:
            print(f"Installing {package}...")
            result = subprocess.run([
                pip_path, "install", package
            ], capture_output=True, text=True)
            
            if result.returncode == 0:
                print(f"Successfully installed {package}")
            else:
                print(f"Failed to install {package}: {result.stderr}")
                return False
        
        return True
    except Exception as e:
        print(f"Error installing packages: {e}")
        return False


system_prompt = """You are helping organize an Express.js project into folders.

I will give you a file’s content and its name. 
You must respond with exactly one category from this list that the file fit into:
[routes, controllers, models, views, middleware, services, utils, config, tests]

If the file does not clearly fit into any category, respond with 'extra' """


import os
import json
from pathlib import Path

from ....parsers.javascript_typescript.jsParser import JsTxParser
from ....parsers.javascript_typescript.handleModuleExport import getModuleExport
from ....parsers.javascript_typescript.js_var_names import get_variable_name_and_type

from .jsDep import getChunkDependencies
from .handle_js_route import getRoutePathAndMethodsOrReturnNone
from .reverseImport import convertToImportStatements, getDepFromImportDict, mergeImportDict

parser = JsTxParser()

async def processEntryFile(file_name,file_dependencies,project_root,packages):

    _,chunks,chunks_name,import_dict = await parser.parse_code(file_name)

    for chunk in chunks:
        deps = None
        skip_chunk = False
        words = chunk.split(".")
        if len(words) > 1:
            route_info = getRoutePathAndMethodsOrReturnNone(chunk,chunks)

            if route_info is not None:
                deps = getChunkDependencies(chunk,chunks_name,False)
                _, dep_to_path_mappings = getExtraDependencies(import_dict)

                for dep in deps:
                    if dep in dep_to_path_mappings.keys():
                        full_path = os.path.normpath(f"{project_root}/{dep_to_path_mappings[dep]}")
                        # print("Full path is ", full_path,full_path.replace(project_root,"").lstrip("/"))

                        if full_path.replace(project_root,"").lstrip("/") not in file_dependencies:
                            skip_chunk = True
                            break

        # is_ignore = len([i for i in deps if dep_to_path_mappings[i] in file_dependencies]) > 0
        if not skip_chunk:
            # print("Processing \n",chunk, "\n","#"*10)

            # feature,entry_file,is_feature_name=True,packages=set(),processed=set(),ignore_paths=None)
            packages, _ = await extractFeatureCode(chunk,file_name,project_root,False,packages,set(),file_dependencies) # extract code

    return packages

def getExtraDependencies(import_deps):
    extra_dependencies = []
    dep_to_path_mappings = {}

    for item in import_deps:
        element = import_deps[item]

        for depItem in element:
            if depItem["alias"] != None:
                deps = [depItem['alias']]
            elif "ext_deps" in depItem.keys() and len(depItem["ext_deps"]) > 0:
                deps = depItem['ext_deps']
            else:
                deps = [depItem['dep']]
            
            extra_dependencies.extend(deps)

            # create dep to path mapping
            for singleDep in deps:
                dep_to_path_mappings[singleDep] = item

    return [extra_dependencies,dep_to_path_mappings]


def getFileNameFromDir(dir,filename):
    if len(dir.strip()) == 0:
        return None
    
    current_files = os.listdir(dir)
    filename = filename.split(".")[0]

    for file in current_files:
        if file.split(".")[0].strip() == filename:
            # print(f"{dir}/{file}", " Found file")
            return f"{dir}/{file}"
    
    return None


async def extractFeatureCode(feature,entry_file,project_root,is_feature_name=True,packages=set(),processed=set(),ignore_paths=None):
    
    # add file to processed
    processed.add(os.path.normpath(entry_file))

    file_root = entry_file.split("/")[:-1]
    file_root = "/".join(file_root)

    _,_,chunks_names,import_deps = await parser.parse_code(entry_file)
    extra_dependencies, dep_to_path_mappings = getExtraDependencies(import_deps)

    chunk_deps = getChunkDependencies(feature,chunks_names,is_feature_name,extra_dependencies)

    # print(chunk_deps,"\n")
    chunk = chunks_names[feature] if is_feature_name else feature
    chunk_imports = {}

    # process deps
    for dep in chunk_deps:

        if dep in dep_to_path_mappings.keys():
            dep_path = dep_to_path_mappings[dep]
            dep_list = chunk_imports[dep_path] if dep_path in chunk_imports.keys() else []

            # print(f"Get dependency {dep} from {dep_path}")

            if ignore_paths is not None and dep_path in ignore_paths:
                continue
            
            dep_list.append(dep)



            # get new dependency file path
            file_path = str(Path(file_root) / Path(dep_path))
            full_file_path = file_path.split("/")
            file_name = full_file_path.pop(-1)

            full_file_path = "/".join(full_file_path)

            full_file_path = getFileNameFromDir(full_file_path,file_name)

            dep_info = None

            if dep_path in import_deps.keys():
                import_info = import_deps[dep_path] 
                dep_info = getDepFromImportDict(import_info,[dep])

            if dep_info is not None:
                root_import = dep_info[0]['standalone'] == True if 'standalone' in dep_info[0].keys() else False
            else:
                root_import = False

            
            # check if dep is a custom file or an external package
            if full_file_path is not None and not root_import:
                packages, processed = await extractFeatureCode(dep,full_file_path,project_root,True,packages,processed)
            elif full_file_path is not None and root_import:
                # print("Manageing root import for ", dep)
                _,deps_chunks,dep_chunks_names,deps_import_deps = await parser.parse_code(full_file_path)

                if dep in dep_chunks_names.keys():
                    packages, processed = await extractFeatureCode(dep,full_file_path,project_root,True,packages,processed) # process normally
                else:
                    # try to find module.export
                    for dep_chunk in deps_chunks:
                        if dep_chunk.strip().startswith("module.export"):
                            packages, processed = await extractFeatureCode(dep_chunk,full_file_path,project_root,False,packages,processed) 
                            break
                
            else:
                # print(f"External Package {dep} Found")
                packages.add(dep_path)

            # print("dependencies list are ", dep_list)
            
            # if deb_object is not None:
            chunk_imports[dep_path] = dep_list

        else:
            # print(f"Get dependency {dep} from current file")
            if dep != feature:
                await extractFeatureCode(dep,entry_file,project_root)

    # print(f"Finished processing {entry_file}\n\n")

    for path in chunk_imports:
        chunk_imports[path] = getDepFromImportDict(import_deps[path],chunk_imports[path])

    entry_file = os.path.normpath(entry_file).replace(project_root,"")

    if entry_file.startswith("/"):
        entry_file = entry_file[1:]

    output_file = os.path.join(project_root,"output")

    # print("#"*20," ",output_file)
    full_path = os.path.join(output_file, entry_file)

    deps_in_file = set()

    print("full filepath is ",full_path)

    if os.path.exists(full_path):
        # add to file
        _,chunks,chunks_names,import_deps = await parser.parse_code(full_path)

        if chunk not in chunks:
            chunks.append(chunk)
            
        mergeImport = mergeImportDict(import_deps,chunk_imports)

        file_imports = convertToImportStatements(mergeImport)
        file_code = chunks

    else:
        # make dir
        os.makedirs(os.path.dirname(full_path),exist_ok=True)
        print("creating folder ", os.path.dirname(full_path))
        file_imports = convertToImportStatements(chunk_imports)
        file_code = [chunk]


    # clean chunks here
    cleaned_chunks = []
    removed_exports = []

    for i in file_code:
        if i.strip().startswith("module.exports") == False:
            cleaned_chunks.append(i)

            # get chunk name here
            chunk_var_name = get_variable_name_and_type(i)
            if chunk_var_name is not None:
                deps_in_file.add(chunk_var_name[1])
        else:
            removed_exports.append(i)

    with open(full_path,"w") as file:
        code = ""
        file_imports = "\n".join(file_imports)

        if len(file_imports) > 0:
            code += f"{file_imports}\n\n"

        # get file export
        exports = await getModuleExport(entry_file,deps_in_file)

        # TODO : this is not standard and it will fail
        # find a better solution.
        """
        so incase we removed an original export from file,
        and after running getModuleExport we were not able to detect any export,
        we just add the original export back.

        this is relevant for database schema where the schema name is defined when exported
        
        e.g : module.exports = mongoose.model("Contact",contactSchema)
        remeber this function is removing every module.exports
        """

        if exports:
            for export in exports:
                if export not in cleaned_chunks:
                    cleaned_chunks.append(export)

        elif removed_exports:
            for entry in removed_exports:
                cleaned_chunks.append(entry)

        file_code = "\n\n".join(cleaned_chunks) # TODO : remove duplicates from chunks

        code += file_code
        file.write(code)

    return packages, processed


def createPackageJsonFile(packages,entry_file,dependencies,dev_dependencies,project_root):
    template_dependencies = {}
    template_dev_dependencies = {}
    data = {}

    for package in packages:
        if package in dependencies.keys():
            template_dependencies[package] = dependencies[package]
        elif package in dependencies.keys():
            template_dev_dependencies[package] = dev_dependencies[package]

    # print("dependencies ", template_dependencies)
    # print("dev dependencies ", template_dev_dependencies)

    # name, main, scripts, description, devDependencies, 
    # data['name'] = "feature_name"
    data['main'] = entry_file
    data['scripts'] = {"dev":f"node {entry_file}"}

    if len(template_dependencies) > 0:
        data['dependencies'] = template_dependencies

    if len(template_dev_dependencies) > 0:
        data['devDependencies'] = template_dev_dependencies

    with open(f"{project_root}/output/package.json","w") as file:
        json.dump(data, file, indent=4)


async def handleExtraction(root_path, feature, file_path):
    packages,files_processed = await extractFeatureCode(feature,file_path,root_path,False)

    package_json = f"{root_path}/package.json"

    if not os.path.exists(f"{root_path}/package.json"):
        raise ValueError("Cannot find package.json in root project folder ",root_path)
    
    with open(package_json,"r") as file:
        data = json.loads(file.read())

        dependencies = data['dependencies'] if "dependencies" in data.keys() else None
        dev_dependencies = data['dev_dependencies'] if "dev_dependencies" in data.keys() else None
        main = f"{root_path}/{data['main']}" if 'main' in data.keys() else None

        while main == None:
            main_file_path = input("Please Enter the full Path to your project entry file : ")
            if os.path.exists(main_file_path):
                main = main_file_path
            else:
                print("The file whose path you specified does not exists")

        # print("Main file is ", main)
        # print("Dependencies are ", dependencies)
        # print("Dev Dependencies are ", dev_dependencies)

        # print("packages ", packages)
        # print("files ", files_processed)

        if main not in files_processed:
            # process main file
            packages = await processEntryFile(main,files_processed,root_path,packages)

        
        # print("Final Packages are ", packages, " type is ", type(packages))

        main_path = main.replace(root_path,"").strip()
        if main_path.startswith("/"):
            main_path = main_path[1:]
        
    createPackageJsonFile(packages,main_path,dependencies,dev_dependencies,root_path)
    return main_path #entry file
        


import threading
import time
import subprocess
import select
import sys


class PythonExecutor:
    def __init__(self, timeout=10):
        self.timeout = timeout
        self.output = [[],[]] #stdout stderror

    def kill_after_timeout(self):
        """Kill process after timeout if still running"""
        time.sleep(self.timeout)
        if not self.stop_event.is_set():
            try:
                if self.process and self.process.poll() is None:
                    print("Terminating due to timeout Jumbo")
                    self.process.terminate()
                    try:
                        self.process.wait(timeout=5)
                    except subprocess.TimeoutExpired:
                        self.process.kill()
                        self.process.wait(timeout=1)
                    finally:
                        self.stop_event.set()
            except Exception as err:
                print(f"Error in timeout handler: {err}")
                self.stop_event.set()

    def executeCommand(self, command,env=None,cwd=None):
        """Execute command and capture output"""
        try:
            self.process = subprocess.Popen(
                command,
                stderr=subprocess.PIPE,stdout=subprocess.PIPE,
                text=True,
                bufsize=1,  # Line buffered
                env=env,
                cwd=cwd,
            )

            # Read output until process finishes or stop event is set
            while not self.stop_event.is_set() and self.process.poll() is None:
                # Check if there's data to read from stdout
                if self.process.stdout:
                    line = self.process.stdout.readline()
                    # print("our line ",line.strip())
                    if line:
                        self.output[0].append(line)
                
                # Check if there's data to read from stderr
                if self.process.stderr:
                    line = self.process.stderr.readline()
                    if line:
                        # print(f"ERROR: {line.strip()}")
                        self.output[1].append(line)
                
                # Small sleep to prevent busy waiting
                time.sleep(0.01)

            # Read any remaining output after process finishes
            if self.process.stdout:
                remaining_stdout = self.process.stdout.read()
                if remaining_stdout:
                    for line in remaining_stdout.splitlines(keepends=True):
                        # print(line.strip())
                        self.output[0].append(line)

            if self.process.stderr:
                remaining_stderr = self.process.stderr.read()
                if remaining_stderr:
                    for line in remaining_stderr.splitlines(keepends=True):
                        # print(f"ERROR bingo : {line.strip()}")
                        self.output[1].append(line)

            print("Process finished")
            self.stop_event.set()

        except Exception as err:
            print(f"Error in executeCommand: {err}")
            self.stop_event.set()
    
    def executeSelfExitingCommand(self,command,env=None,cwd=None):

        process = subprocess.Popen(command,env=env,cwd=cwd,stdout=subprocess.PIPE,stderr=subprocess.PIPE)
        
        stdout, stderr = process.communicate()

        if stdout:
            lines = stdout.decode()
            lines = lines.split("\n")
            for line in lines:
                if len(line.strip()) > 0:
                    self.output[0].append(line)

        if stderr:
            lines = stderr.decode()
            lines = lines.split("\n")
            for line in lines:
                if len(line.strip()) > 0:
                    self.output[1].append(line)

    def runCommand(self, command, self_exit=False,env=None,cwd=None):
        """Run command with timeout monitoring"""
        self.output = [[],[]] #[output, error]
        self.process = None

        if self_exit:
            self.executeSelfExitingCommand(command,env,cwd)
        else:
            self.stop_event = threading.Event()
            
            # Start execution thread
            exec_thread = threading.Thread(target=self.executeCommand, args=(command,env,cwd,))
            timeout_thread = threading.Thread(target=self.kill_after_timeout)
            
            exec_thread.start()
            timeout_thread.start()

            # Wait for execution to complete
            exec_thread.join()
            
            # Stop timeout thread if execution finished first
            self.stop_event.set()
            
            # Don't wait too long for timeout thread
            timeout_thread.join(timeout=1)

        return self.output


# Alternative implementation using threading for concurrent stdout/stderr reading
class ImprovedPythonExecutor:
    def __init__(self, timeout=10):
        self.timeout = timeout
        self.output = [[],[]]

        self.stop_event = threading.Event()
        self.process = None

    def read_stream(self, stream, stream_name):
        """Read from a stream (stdout or stderr) in a separate thread"""
        try:
            for line in iter(stream.readline, ''):
                if self.stop_event.is_set():
                    break
                if line:
                    # if stream_name == 'stderr':
                    #     print(f"ERROR: {line.strip()}")
                    # else:
                    #     print(line.strip())
                    if stream_name == "stdout":
                        self.output[0].append(line)
                    else:
                        self.output[1].append(line)

                    # self.output.append((stream_name, line))
        except Exception as err:
            print(f"Error reading {stream_name}: {err}")
        finally:
            stream.close()

    def timeout_handler(self):
        """Handle timeout"""
        time.sleep(self.timeout)
        if not self.stop_event.is_set():
            print("Terminating due to timeout")
            self.stop_event.set()

            if self.process and self.process.poll() is None:
                self.process.terminate()
                try:
                    self.process.wait(timeout=5)
                except subprocess.TimeoutExpired:
                    self.process.kill()

    def runCommand(self, command,cwd):
        """Run command with improved concurrent output handling"""
        self.output = [[],[]]
        self.stop_event = threading.Event()

        try:
            # Start process
            self.process = subprocess.Popen(
                command,
                cwd=cwd,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                text=True,
                bufsize=1
            )

            # Start reader threads
            stdout_thread = threading.Thread(
                target=self.read_stream, 
                args=(self.process.stdout, 'stdout'),
                daemon=True
            )
            stderr_thread = threading.Thread(
                target=self.read_stream, 
                args=(self.process.stderr, 'stderr'),
                daemon=True
            )
            timeout_thread = threading.Thread(target=self.timeout_handler,daemon=True)

            stdout_thread.start()
            stderr_thread.start()
            timeout_thread.start()

            # Wait for process to complete
            self.process.wait()
            
            # Signal threads to stop
            self.stop_event.set()

            # Wait for threads to finish
            stdout_thread.join(timeout=2)
            stderr_thread.join(timeout=2)
            timeout_thread.join(timeout=1)

            print("Command completed")

        except Exception as err:
            # print(f"Error: {err}")
            self.stop_event.set()

        return self.output


# # Test both implementations
# if __name__ == "__main__":
#     print("=== Testing Original Fixed Version ===")
#     executor1 = PythonExecutor(timeout=5)
#     result1_output, result1_error = executor1.runCommand(["pip", "install","django-boto"],True)
#     print(f"Captured {len(result1_output)} lines")
    
#     print("\n","#"*10,"\n")
#     print("\n".join(result1_output))

#     result2, result2_error = executor1.runCommand(["python", "manage.py","runserver"])
#     print("\n Result 2 \n")
#     print("".join(result2_error))
    
#     # print("\n=== Testing Improved Version ===")
#     # executor2 = ImprovedPythonExecutor(timeout=5)
#     # result2 = executor2.runCommand(["python", "manage.py","runserver"])
#     # print(f"Captured {len(result2)} lines")
    
#     # print("\n=== Testing with a command that has both stdout and stderr ===")
#     # executor3 = ImprovedPythonExecutor(timeout=5)
#     # result3 = executor3.runCommand(["python3", "-c", "import sys; print('stdout message'); print('stderr message', file=sys.stderr)"])
#     # print(f"Captured {len(result3)} lines")
#     # for stream, line in result3:
#     #     print(f"{stream}: {line.strip()}")


prompt = """
You are an intelligent prompt parser for a Django application customization service.

Your task is to analyze the user's request and break it down into a list of clear, 
actionable sub-prompts.

Do not solve or execute any sub-prompt!

Carefully read the user’s input and identify each distinct action they are requesting. 
Break these into atomic, actionable sub-prompts that another agent can execute one at a time.

Your output should:

*Be a plain, numbered list of sub-prompts
*Be specific to Django (e.g., models, views, templates, forms, admin, settings, etc.)
*Use clear, unambiguous language
*Not skip or combine tasks; each item should represent one actionable task

NOTE!! : Do not include explanations, introductions, or summaries. Only output the list.

Example Input from User:

add purchased_by to the order models, 
edit the checkout function to send a notifiction email on successful payment and
also register all models in the admin panel

Example Output:

Add a purchased_by field to the Order model.
Edit the checkout function to send a notification email upon successful payment.
Register all models in the Django admin panel.

"""


import asyncio
import os
from pathlib import Path
import subprocess

from .js_var_names import get_variable_name_and_type

from ...src.Express.extract.jsDep import getLineValue


"""
- variable name
- recursively extract dependencies from source code 
- extract package dependencies from package.json
"""

"""
check if u can split if statements, sitch statements too
"""

# TODO : how are we identifying file names

output_marker = "### PARSER OUTPUT ###"
varIdentifies = ["const ","let ","var "]

class JsTxParser:
    def __init__(self):
        self.opening = [ "[","(","{","/*"]
        self.closing = ["]",")","}","*/"]
        self.cache = None

    def get_and_set_variable_name(self,chunk):

        return get_variable_name_and_type(chunk)
    
        # TODO : implement connection to node process that use AST to get variable name
        command = ["npm", "run", "dev", chunk]

        process = subprocess.Popen(
            command,
            stdin=subprocess.PIPE,
            stderr=subprocess.PIPE,
            stdout=subprocess.PIPE,
            text=True,
            bufsize=1  # Line buffered
        )

        stdout, stderr = process.communicate()

        if stdout:
            response = stdout.split(output_marker)[-1].strip()
            print(response, " name of chunk")
            if len(response) > 0:
                var_type, var_name = response.split("-")
                # print([var_type, var_name])
                return [var_type.strip(), var_name.strip()]
        

        if stderr:
            print("error ",stderr)

        return None
    
    def getFileInfoFromCache(self,filename):
        if self.cache is not None and filename in self.cache.keys():
            return self.cache[filename]
        return None

    def storeFileInfoInCache(self,filename, fileData):
        # TODO : find out; is this optimized
        cache = self.cache if self.cache is not None else {}
        cache[filename] = fileData

        self.cache = cache

    def clearCache(self):
        self.cache = None

    async def parse_code(self,filename, ignore_import=False, ignore_chunk_name=False):

        # check cache if file have been processed before
        cache_file_data = self.getFileInfoFromCache(filename)

        if cache_file_data is not None:
            return cache_file_data

        stack = []
        memory = []

        chunks = []
        imports = []

        variableToChunkMapping = {}

        with open(filename,"r") as file:
            code = file.read()

            lines = code.split("\n")
            for line in lines:
                is_import = False
                memory.append(line)

                # handle multi line comments
                if line.startswith("/*") and line.endswith("*/"):
                    pass # skip edge case where multi line comment are on 1 line. E.g : - /*Hello my Guy We land shjgsgs shgshgs*/ -

                elif line.startswith("/*"):
                    # print("multi line comment opening found")
                    stack.append("/*")

                elif line.endswith("*/"):
                    # print("multi line comment closing found")
                    if len(stack) > 0:
                        last = stack[-1]
                        index = self.opening.index(last)

                        if index == self.closing.index("*/"):
                            stack.pop(-1)

                    else:
                        raise ValueError("Your code has an error and its not syntactic correct")
                    
                else:
                    for char in line:
                        if char in self.opening:
                            stack.append(char)
                        elif char in self.closing:
                            if len(stack) > 0:
                                last = stack[-1]
                                index = self.opening.index(last)

                                if index == self.closing.index(char):
                                    stack.pop(-1)

                            else:
                                raise ValueError("Your code has an error and its not syntactic correct")

                if len(stack) == 0:
                    newCode = "\n".join(memory)
                    if len(newCode.strip()) > 0:
                        # TODO check if newCode start with special characters like ||, else, else if ....
                        if newCode.strip()[0] in self.opening and len(chunks) > 0:
                            last_chunk = chunks.pop(-1)
                            newCode = f"{last_chunk}\n{newCode}"

                        if not ignore_import and newCode.strip().startswith("import "):
                            imports.append(newCode)
                            is_import = True
                        else:
                            chunks.append(newCode)

                            if not newCode.startswith("/") and not ignore_chunk_name:

                                if newCode.strip().startswith("exports."):
                                    chunk_to_process = "const "+newCode.strip()[8:]
                                else:
                                    chunk_to_process = newCode

                                # print(chunk_to_process,"\n\n")
                                chunk_info = self.get_and_set_variable_name(chunk_to_process)
                                # print(chunk_info, " - chunk info")
                                if chunk_info is not None:
                                    if isinstance(chunk_info,list):
                                        # print("info ",chunk_info[1])
                                        variableToChunkMapping[chunk_info[1]] = newCode
                                        # print("mapping ", variableToChunkMapping)
                                else:
                                    # print("chunk evaluated to none ",chunk_to_process)
                                    pass

                                # store_variable_name(newCode)
                            # chunk_info = self.get_and_set_variable_name(newCode)
                            # if chunk_info is not None:
                            #     if isinstance(chunk_info,list):
                            #         variableToChunkMapping[chunk_info[1]] = newCode

                    memory = []


            if len(stack) > 0:
                raise ValueError("Your code has an error and its not syntactic correct")
            
        import_deps = {}
        require_synthax_import = []
        new_chunks = []

        for chunk in chunks:
            require_check = "require(" in chunk

            # convert require to import syntax
            if require_check:
                chunk = convertRequireToImport(chunk)
                imports.append(chunk)
                require_synthax_import.append(chunk)

                # TODO: remove chunk from chunks
                # chunks.remove(chunk)
            else:
                new_chunks.append(chunk)

        chunks = new_chunks

        for chunk in imports:
            # stand_alone_deps = None # that is dependencies inside { } e.g {deps1, deps2} example import statement import {deps} from "path"

            if " from " in chunk:
                deps, src = chunk.split(" from ")

                namedImports = []

                if "{" in deps and "}" in deps:
                    namedImports = deps.split("}")[0].split("{")[-1].split(",")


                deps = deps.replace("{","").replace("}",'').replace("import","").strip().split(",")
                src = src.replace("'","").replace('"',"")

                if src.endswith(";"):
                    src = src[:len(src)-1]

                src_deps = []

                if src in import_deps.keys():
                    src_deps = import_deps[src]

                
                for word in deps:
                    if " as " in word:
                        word, alias = word.split(" as ")
                        src_deps.append({"dep":word.strip(),"alias":alias.strip(),"standalone": word not in namedImports, "use-require":chunk in require_synthax_import})
                    
                    elif word.strip() == "*":
                        # print("getting chunks ", src)

                        # Example paths
                        path1 = Path(os.path.dirname(filename))
                        path2 = Path(f"{src}.js")

                        merged = path1 / path2

                        _, chunks, varMappings = self.parse_code(merged,True)
        
                        src_deps[-1]['ext_deps'] = list(varMappings.keys())
                        skip_postprocessing = True

                        for ext_dep in list(varMappings.keys()):
                            src_deps.append({"dep":ext_dep.strip(),"alias":None,"standalone":True,"use-require":chunk in require_synthax_import})

                    else:
                        src_deps.append({"dep":word.strip(),"alias":None,"standalone": word not in namedImports, "use-require":chunk in require_synthax_import})

                    import_deps[src] = src_deps
            else:
                pass
                # print("no from in chunk ",chunk,"\n")
        
        returnObject = [imports,chunks,variableToChunkMapping,import_deps]

        # Store file data in cache
        self.storeFileInfoInCache(filename,returnObject)

        return returnObject 

def removeJsIndentifier(line):
    for indentify in varIdentifies:
        if line.startswith(indentify):
            return line.replace(indentify,"").strip()
    
    return None


def convertRequireToImport(line):
    lineData = getLineValue(line)
    if lineData is not None:
        beforeEqualSign, AfterEqualSign = lineData

        # get path
        end = -1
        start = 0

        while AfterEqualSign[start] != "(":
            start += 1

        while AfterEqualSign[end] != ")":
            end -= 1

        path = AfterEqualSign[start+1:end]
        path = path.replace("'","").replace('"',"")

        # get dependencies
        deps = removeJsIndentifier(beforeEqualSign[:-1])

        if " :" in deps:
            deps = deps.replace(" :","as")
        elif ":" in deps:
            deps = deps.replace(":"," as")

        return f'import {deps} from "{path}"'


# if __name__ == "__main__":
#     parser = JsTxParser()

#     print("Parsing Test.js")
#     imports, chunks, chunks_name, import_deps = asyncio.run(parser.parse_code("test.js"))

#     print("names ",chunks_name)
#     print(chunks)

    # import_deps = {}

    # for chunk in chunks:
    #     require_check = "require(" in chunk

    #     # convert require to import syntax
    #     if require_check:
    #         chunk = convertRequireToImport(chunk)
    #         imports.append(chunk)

    #         # TODO: remove chunk from chunks
    #         # chunks.remove(chunk)

    # for chunk in imports:
    #     if " from " in chunk:
    #         deps, src = chunk.split(" from ")
    #         deps = deps.replace("{","").replace("}",'').replace("import","").strip().split(",")

    #         src = src.replace("'","").replace('"',"")

    #         if src.endswith(";"):
    #             src = src[:len(src)-1]

    #         src_deps = []

    #         if src in import_deps.keys():
    #             src_deps = import_deps[src]

            
    #         for word in deps:
    #             if " as " in word:
    #                 word, alias = word.split(" as ")
    #                 src_deps.append({"dep":word.strip(),"alias":alias.strip()})
    #             else:
    #                 src_deps.append({"dep":word.strip(),"alias":None})
                
    #             if word.strip() == "*":
    #                 print("getting chunks ", src)

    #                 # Example paths
    #                 path1 = Path("express.js_contact_app/controllers/")
    #                 path2 = Path(f"{src}.js")

    #                 merged = path1 / path2

    #                 _, chunks, varMappings = parser.parse_code(merged,True)
    #                 # print("external deps ",chunks)
    #                 print(varMappings.keys())
    #                 src_deps[-1]['ext_deps'] = list(varMappings.keys())

    #             import_deps[src] = src_deps
    #     else:
    #         print("no from in chunk ",chunk,"\n")
    
#     print("import dictionary ", json.dumps(import_deps,indent=4))
#     print("unique chunk count ", len(chunks))

#     chunk = """const getContacts= asyncHandler(async (req,res)=>{
#     const hello = 1
#     const contacts = await Contact.find({user_id:req.user._id})
#     const name = hello
#     res.status(200).json(contacts)
# })"""

#     chunk_dep = getChunkDependencies("updateContact",chunks_name)
#     print("deps = ", chunk_dep)


import os
import esprima

from ....parsers.javascript_typescript.jsParser import JsTxParser
from ....src.Express.extract.handle_js_route import getRoutePathAndMethodsOrReturnNone
from ....src.Express.extract.jsDep import getChunkDependencies
from ....src.Express.extract.move_file import getNewImportPath

from rich.prompt import Prompt


parser = JsTxParser()

def get_express_instance_variable(js_code: str):
    tree = esprima.parseScript(js_code)

    express_var = None
    for node in tree.body:
        # Look for: const app = express();
        if node.type == "VariableDeclaration":
            for decl in node.declarations:
                if (
                    decl.init
                    and decl.init.type == "CallExpression"
                    and decl.init.callee.type == "Identifier"
                    and decl.init.callee.name == "express"
                ):
                    express_var = decl.id.name  # e.g., "app"
    return express_var

def getFileContent(filename):
    with open(filename,"r") as file:
        return file.read()
    return None

def cleanupChunk(chunk,template_express_var,destination_express_var):

    if chunk.strip().startswith(template_express_var):
        chunk = chunk.split(template_express_var)[1:]
        chunk.insert(0,destination_express_var)
        return "".join(chunk)
    return chunk

def findImportStatement(dependency,import_deps):
    for import_path in import_deps:
        dependcies = import_deps[import_path]

        # TODO : get actual import path by crawling project folders

        for i in dependcies:
            if i['dep'] == dependency:
                return [import_path,dependcies]
            
    return None

def processDependencies(
        entry_file_path,
        template_files,
        chunk,
        chunks,
        chunks_name,
        target_listener,
        template_express_var,
        destination_express_var,
        target_chunks,
        import_deps,
        target_import_deps
    ):

    deps = getChunkDependencies(chunk,chunks_name,False)
    for dep in deps:
        
        dep_chunk = chunks_name[dep]

        dep_import = findImportStatement(dep,import_deps)
        if  dep_import is not None:
            path,dep_mapping = dep_import

            cleaned_path = path

            while cleaned_path.startswith("."):
                cleaned_path = cleaned_path[1:]
                if len(cleaned_path) == 1 and cleaned_path == ".":
                    raise ValueError("Invalid path")
                
            for filename in template_files:

                if filename.endswith(cleaned_path):
                    import_path = getNewImportPath(entry_file_path,filename)
                    
                    if not import_path.startswith("."):
                        import_path = f"./{import_path}"

                    target_import_deps[import_path] = dep_mapping
                    break
    
        if dep_chunk not in target_chunks and dep_chunk in chunks:
            insert_index = target_chunks.index(target_listener)
            insert_chunk = cleanupChunk(dep_chunk,template_express_var,destination_express_var)

            if insert_chunk not in target_chunks:
                target_chunks.insert(insert_index,insert_chunk)
            
            # recursive with dep_chunk
            target_chunks,import_deps = processDependencies(entry_file_path,template_files,dep_chunk,chunks,chunks_name,target_listener,template_express_var,destination_express_var,target_chunks,import_deps,target_import_deps)

    insert_index = target_chunks.index(target_listener)
    insert_chunk = cleanupChunk(chunk,template_express_var,destination_express_var)
    
    if insert_chunk not in target_chunks:
        target_chunks.insert(insert_index,insert_chunk)
    
    return target_chunks,target_import_deps

# what are you doing?
async def mergeEntryFiles(project_entry, template_entry,template_files=None,logger=None):

    logger.start_status("Processing project entry file")

    entry_file = os.path.dirname(project_entry)

    first = getFileContent(project_entry)
    template = getFileContent(template_entry)

    express_var_name_first = get_express_instance_variable(first)
    express_var_name_second = get_express_instance_variable(template)

    template = template.replace(f"{express_var_name_second}",express_var_name_first)

    _,chunks,chunks_name,import_deps = await parser.parse_code(template_entry)
    

    _,target_chunks,_,target_import_deps = await parser.parse_code(project_entry)
    target_listener = None

    for i in target_chunks:
        if i.strip().startswith(f"{express_var_name_first}.listen"):
            target_listener = i

    try:
        chunks.remove(chunks_name[express_var_name_second])
    except:
        pass

    selected_chunks = []
    for chunk in chunks:
        if chunk.strip().startswith(f"{express_var_name_second}.listen"):
            deps = getChunkDependencies(chunk,chunks_name,raw_code=False)
            for dep in deps:
                if dep in chunks_name.keys():
                    if chunks_name[dep] in selected_chunks:
                        selected_chunks.remove(chunks_name[dep])
        else:
            if chunk.startswith("/") == False:
                selected_chunks.append(chunk)

    logger.stop_status()

    for chunk in selected_chunks:
        info = getRoutePathAndMethodsOrReturnNone(chunk,None)
        if info is not None:
           target_chunks,target_import_deps = processDependencies(
                entry_file,
                template_files,
                chunk,
                chunks,
                chunks_name,
                target_listener,
                express_var_name_second,
                express_var_name_first,
                target_chunks,
                import_deps,
                target_import_deps
            )
        else:
            res = Prompt.ask(f"[cyan]Should I add \n{chunk} to entry file (yes/no) [/cyan]")
            while res.lower() not in ["yes","no"]:
                res = Prompt.ask(f"[cyan]Please enter a valid response (yes/no) [/cyan]")

            if res.lower() == "no":
                continue

            target_chunks,target_import_deps = processDependencies(
                entry_file,
                template_files,
                chunk,
                chunks,
                chunks_name,
                target_listener,
                express_var_name_second,
                express_var_name_first,
                target_chunks,
                import_deps,
                target_import_deps
            )

    return target_import_deps, target_chunks


